from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.complexity.word.name_entity_enum import NamedEntityONEnum 
from rb.core.text_element_type import TextElementType
from typing import List, Callable, Set
from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class NoUnqNamedEntity(ComplexityIndex):

    """named_ent_type depends on the model, there are 2 enum options
        see name_entity_enum.py for those"""
    def __init__(self, lang: Lang, named_ent_type: NamedEntityONEnum,
            reduce_depth: int, reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.WORD,
                                 abbr="UnqNmdEnt", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)
        self.named_ent_type = named_ent_type

    def _compute_value(self, element: TextElement) -> int:
        return len({word.text for word in element.get_words() if word.ent_type_ != 0 and word.ent_type_ == self.named_ent_type.name})
    
    def __repr__(self):
        return f"{self.reduce_function_abbr}({self.abbr}_{self.named_ent_type.name.lower()} / {self.reduce_depth_abbr})"


from rb.complexity.complexity_index import ComplexityIndex
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.complexity.rhythm.syllabified_dict import SyllabifiedDict
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.utils.rblogger import Logger

from typing import List

logger = Logger.get_logger()


class NoAlliterations(ComplexityIndex):

    def __init__(self, lang: Lang, reduce_depth: int,
                 reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.RHYTHM,
                                 abbr="RhNoAlliterations", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)

        self.syllabified_dict = SyllabifiedDict.get_instance(lang)

    def process(self, element: TextElement) -> float:
        pass




from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.complexity_index import ComplexityIndex
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from typing import Set
from rb.core.text_element_type import TextElementType
from typing import Callable, List, Set
from rb.utils.rblogger import Logger

logger = Logger.get_logger()

class NoUniqueWordsIndex(ComplexityIndex):
    

    def __init__(self, lang: Lang, reduce_depth: int,
                 reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.SURFACE,
                                 abbr="UnqWd", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)

    def _compute_value(self, element: TextElement) -> int:
        return len({word.text for word in element.get_words()})


import unittest
from rb.core.document import Document
from rb.core.lang import Lang
from rb.similarity.word2vec import Word2Vec
from rb.similarity.lda import LDA
from rb.similarity.lsa import LSA
from rb.similarity.aoa import AgeOfAcquisition

class SimilarityTest(unittest.TestCase):

    def test_load_word2vec(self):
        w2v = Word2Vec('coca', Lang.EN)
        self.assertEqual(len(list(w2v.vectors.values())[0]), 300, "The dimension should have been 300")

    def test_load_lda(self):
        lda = LDA('coca', Lang.EN)
        self.assertEqual(len(list(lda.vectors.values())[0]), 300, "The dimension should have been 300")

    def test_load_lsa(self):
        lsa = LSA('coca', Lang.EN)
        self.assertEqual(len(list(lsa.vectors.values())[0]), 300, "The dimension should have been 300")

    def test_load_aoa(self):
        aoa = AgeOfAcquisition(Lang.EN)
        kuperman = aoa.get_kuperman_value("antihuman")
        self.assertEqual(kuperman, 11.08, "The kuperman value of 'antihuman' should have been 11.08")
        shock = aoa.get_shock_value("argue")
        self.assertEqual(shock, 4.24, "The shock value of 'argue' should have been 4.24")
        bird = aoa.get_bird_value("argument")
        self.assertIsNone(bird, "The bird value for 'argument' should have been None")

if __name__ == '__main__':
    unittest.main()


import re
from typing import Dict, Iterable, List, Union

import spacy
from nltk.tokenize import sent_tokenize
from rb.core.lang import Lang
from rb.core.pos import POS
from rb.utils.rblogger import Logger
from spacy.language import Language
from spacy.tokens import Doc, Span, Token

logger = Logger.get_logger()

models = {
    Lang.EN: 'en_core_web_lg',
    Lang.NL: 'nl_core_news_lg',
    Lang.FR: 'fr_core_news_lg',
    Lang.ES: 'es_core_news_lg',
    Lang.DE: 'de_core_news_lg',
    Lang.IT: 'it_core_news_lg',
    Lang.RO: 'ro_core_news_lg',
    Lang.RU: 'ru_core_news_lg',
}

normalization = {
    'ro': [
        (re.compile("ş"), "ș"),
        (re.compile("Ş"), "Ș"),
        (re.compile("ţ"), "ț"),
        (re.compile("Ţ"), "Ț"),
        (re.compile("(\w)î(\w)"), "\g<1>â\g<2>")
    ]
}

re_missing_space = re.compile("([a-z]+)\.([A-Z][a-z]+)")

def convertToPenn(pos: str, lang: Lang) -> str:
    if lang == Lang.FR:
        pos = pos.lower()
        if pos.startswith('noun') or pos.startswith('propn'):
            return "NN"
        if pos.startswith("verb"):
            return "VB"
        if pos.startswith("adj"):
            return "JJ"
        if pos.startswith("adv"):
            return "RB"
        if pos.startswith("adp"):
            return "IN"
        if pos.startswith("cconj"):
            return "CC"
        return ""
    if lang == Lang.NL:
        pos = pos.lower()
        if pos.startswith('n_') or pos.startswith('n|') or pos.startswith('propn'):
            return "NN"
        if pos.startswith("v_") or pos.startswith("v|"):
            return "VB"
        if pos.startswith("adj"):
            return "JJ"
        if pos.startswith("adv"):
            return "RB"
        if pos.startswith("adp"):
            return "IN"
        if pos.startswith("cconj") or pos.startswith("conj"):
            return "CC"
        return ""
    if lang == Lang.RO: 
        pos = pos.lower()
        if pos.startswith("n"):
            return "NN"
        if pos.startswith("v"):
            return "VB"
        if pos.startswith("a"):
            return "JJ"
        if pos.startswith("r"):
            return "RB"
        if pos.startswith("s") or pos.startswith("cs"):
            return "IN"
        if pos.startswith("c"):
            return "CC"
        return ""
    if len(pos) > 2:
        return pos[:2]
    return pos
    
def computePOS(token: Union[str, Token], lang: Lang) -> POS:
    if lang == Lang.RO:
        if isinstance(token, Token):
            pos = token.tag_.lower()
        else:
            pos = token.lower()
        if pos.startswith("comma") or pos.startswith("period"):
            return POS.PUNCT
        if pos.startswith("n"):
            return POS.NOUN
        if pos.startswith("v"):
            return POS.VERB
        if pos.startswith("a"):
            return POS.ADJ
        if pos.startswith("r"):
            return POS.ADV
        if pos.startswith("s") or pos.startswith("cs"):
            return POS.ADP
        if pos.startswith("c"):
            return POS.CCONJ
        if pos.startswith("m"):
            return POS.NUM
        if pos.startswith("t"):
            return POS.DET
        if pos.startswith("d"):
            return POS.PRON
        return POS.X
    try:
        return POS(token.pos_)
    except:
        return POS.X

class SpacyParser:

    _INSTANCE = None

    def __init__(self):
        self.pipelines = {
            lang: spacy.util.get_lang_class(lang.value)()
            for lang in models
        }
        # for pipeline in self.pipelines.values():
        #     component = pipeline.create_pipe('tagger')   # 3. create the pipeline components
        #     pipeline.add_pipe(component)
        self.loaded_models = {}
    
    @classmethod
    def get_instance(cls) -> "SpacyParser":
        if cls._INSTANCE is None:
            cls._INSTANCE = SpacyParser()
        return cls._INSTANCE
        
    def preprocess(self, text: str, lang: str) -> str:
        if lang not in normalization:
            return text
        for pattern, replacement in normalization[lang]:
            text = re.sub(pattern, replacement, text)
        return text

    def get_tokens_lemmas(self, sentences: Iterable, lang: str) -> Iterable:
        if lang not in self.pipelines:
            return None
        pipeline = self.pipelines[lang]
        # sbd = pipeline.create_pipe('sentencizer')
        # pipeline.add_pipe(sbd)
        doc = pipeline.pipe((sent[:1].lower() + sent[1:] for sent in sentences), batch_size=100000, n_threads=16)
        # print([sent.string.strip() for sent in doc.sents])
        # print(len(doc.sents))
        # print("====================")
        # for token in doc:
        #     print(token.text, token.lemma_, token.pos_, token.tag_, token.dep_,
        #   token.shape_, token.is_alpha, token.is_stop)
        # print("====================")
        return doc
        # return [(token.text, token.lemma_) for token in doc]

    def parse_block(self, block: str, lang: Lang) -> List[Span]:
        block = re.sub(re_missing_space, r"\1. \2", block)
        doc = self.parse(block, lang)
        return [sent for sent in doc.sents]

    def tokenize_sentences(self, block: str) -> List[str]:
        return sent_tokenize(block)
    
    @staticmethod
    @Language.component('line_splitter')
    def line_splitter(doc):
        for token in doc:
            if "\n" in token.text_with_ws and token.i < len(doc) - 1:
                token.nbor().sent_start = True
        return doc


    def get_model(self, lang: Union[Lang, str]) -> Language:
        if isinstance(lang, str):
            lang = Lang(lang)
        if lang not in self.loaded_models:
            try:
                self.loaded_models[lang] = spacy.load(models[lang])
                self.loaded_models[lang].add_pipe("line_splitter", name='sentence_segmenter', before='parser')
            except:
                logger.error("spaCy model not found. You can download it by running the following commmand:\npython3 -m spacy download {}".format(models[lang]))
                return None
        return self.loaded_models[lang]

    def parse(self, sentence: str, lang: Lang) -> Doc:
        model = self.get_model(lang)
        if model is None:
            raise Exception("Model not loaded")
        doc = model(sentence)
        for token in doc:
            token.pos_ = computePOS(token, lang).value
        return doc


    def is_dict_word(self, word: str, lang: Union[Lang, str]) -> bool:
        model = self.get_model(lang)
        return model.vocab.has_vector(word)

if __name__ == "__main__":
    spacyInstance = SpacyParser()

    sent = """
        După terminarea oficială a celui de-al doilea război mondial, în conformitate cu discursul lui W. Churchill (prim ministru al Regatului Unit la acea dată), de la Fulton, s-a declanșat Războiul rece și a apărut conceptul de cortină de fier. Urmare a politicii consecvente de apărare a sistemului economic și politic (implicit a intereslor economice ale marelui capital din lumea occidentală) trupele germane, în calitate de "prizonieri", aflate pe teritoriul Germaniei de Vest au fost reînarmate și au constituit baza viitorului "Bundeswehr" - armata regulată a R.F.G.

        Pe fondul evenimentelor din 1948 din Cehoslovacia (expulzări ale etnicilor germani, alegeri, reconstrucție economică) apare infiltrarea agenților serviciilor speciale ale S.U.A. și Marii Britanii cu rol de "agitatori". Existând cauza, trupele sovietice nu părăsesc Europa Centrală și de Est cucerită-eliberată, staționând pe teritoriul mai multor state. Aflate pe linia de demarcație dintre cele două blocuri foste aliate, armata sovietică nu a plecat din Ungaria decât după dizolvarea Tratatului de la Varșovia.
        """

    # sent = """
    #     După terminarea oficială a celui de-al doilea război mondial, în conformitate cu discursul lui Churchill, de la Fulton, s-a declanșat Războiul rece și a apărut conceptul de cortină de fier."""

    # print(spacyInstance.get_tokens_lemmas(sent))
    # doc = spacyInstance.parse("My sister has a dog. She loves him.", 'en')
    doc = spacyInstance.parse("Pensée des enseignants, production d’écrits, ingénierie éducative, enseignement à distance, traitement automatique de la langue, outils cognitifs, feedback automatique", 'fr')
    for token in doc:
        print(convertToPenn(token.tag_, 'fr'))   
    # print(spacyInstance.preprocess("coborî", 'ro'))



from statistics import mean
from typing import List, Callable, Dict
import csv
from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.complexity.word.aoa_enum import AoaTypeEnum
from rb.core.text_element_type import TextElementType
from rb.utils.rblogger import Logger

logger = Logger.get_logger()

class Aoa(ComplexityIndex):


    aoa_dict: Dict[AoaTypeEnum, Dict[str, float]] = None

    def __init__(self, lang: Lang, aoa_type: AoaTypeEnum,
            reduce_depth: int, reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.WORD,
                                 abbr="AoA", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)
        if Aoa.aoa_dict is None:
            Aoa.parse_aoa_csv(lang)
        self.aoa_type = aoa_type

    @staticmethod
    def parse_aoa_csv(lang: Lang):

        if lang is Lang.EN:
            path_to_aoa = 'resources/en/wordlists/AoA.csv'
        Aoa.aoa_dict = {}
        for at in AoaTypeEnum:
            Aoa.aoa_dict[at] = {}

        with open(path_to_aoa) as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=',')
            for i, row in enumerate(csv_reader):
                if i < 2:
                    continue
                row[0] = row[0].strip()
                if len(str(row[1])) > 0:
                    Aoa.aoa_dict[AoaTypeEnum.BIRD][row[0]] = float(row[1])
                if len(str(row[2])) > 0:
                    Aoa.aoa_dict[AoaTypeEnum.BRISTOL][row[0]] = float(row[2])
                if len(str(row[3])) > 0:
                    Aoa.aoa_dict[AoaTypeEnum.CORTESE][row[0]] = float(row[3])
                if len(str(row[4])) > 0:
                    Aoa.aoa_dict[AoaTypeEnum.KUPERMAN][row[0]] = float(row[4])
                if len(str(row[5])) > 0:
                    Aoa.aoa_dict[AoaTypeEnum.SHOCK][row[0]] = float(row[5])

    def _compute_value(self, element: TextElement) -> float:
        values = [Aoa.aoa_dict[self.aoa_type][word.text] for word in element.get_words() if word.text in Aoa.aoa_dict[self.aoa_type]]
        return mean(values) if values else 0
    
    def __repr__(self):
        return f"{self.reduce_function_abbr}({self.abbr}_{self.aoa_type.name.lower()} / {self.reduce_depth_abbr})"


from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType   
from typing import List, Callable
from rb.core.pos import POS as PosEum
from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class PosMain(ComplexityIndex):

    
    def __init__(self, lang: Lang, pos_type: PosEum,
            reduce_depth: int, reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.MORPHOLOGY,
                                 abbr="POS", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)
        self.pos_type = pos_type

    def _compute_value(self, element: TextElement) -> int:
        return sum(1 for word in element.get_words() if word.pos is self.pos_type)
    
    def __repr__(self):
        return f"{self.reduce_function_abbr}({self.abbr}_{self.pos_type.name.lower()} / {self.reduce_depth_abbr})"



from rb.comprehension.utils.graph.cm_edge_type import CmEdgeType
from rb.comprehension.utils.distance_strategies.abstract_distance_strategy import DistanceStrategy

from rb.core.word import Word

from rb.similarity.vector_model import VectorModel

from typing import List
Models = List[VectorModel] 


class FullSemanticSpaceWordDistanceStrategy(DistanceStrategy):

    def __init__(self, semantic_models: Models, threshold: float):
        self.semantic_models = semantic_models
        self.threshold = threshold


    def get_distance(self, first_word: Word, second_word: Word) -> float:
        if not self.semantic_models:
            return 0.0

        avg = 0.0
        for sm in self.semantic_models:
            avg += sm.similarity(first_word, second_word)
        avg /= len(sm)

        if avg > self.threshold:
            return avg
        return 0.0

    
    def get_edge_type(self) -> CmEdgeType:
        return CmEdgeType.Semantic


import json
import sys
# import matplotlib.pyplot as plt
import copy
import numpy as np
import tensorflow as tf
from sklearn.model_selection import StratifiedShuffleSplit
from sklearn.utils import class_weight
from collections import Counter
import random
from tensorflow.keras.callbacks import Callback
from sklearn.metrics import classification_report
from sklearn.metrics import accuracy_score
from nltk.tokenize import sent_tokenize
import os

# read entire json file
# if loading the original dataset ignore the reviews with score 0
# return a list of json entries
def readJson(file_path, original=False):
    data = []
    with open(file_path, encoding="utf8") as json_file:
        for line in json_file:
            entry = json.loads(line)
            if original == True:
                if entry['_source']['Review']['ReviewRating'] == 0:
                    continue
            data.append(entry)
    return data


# compute histogram of review scores
# input -> list of jsons
# output -> dict score -> #reviews
def computeScoreHistogram(data, normalize = False):
    histo = {}

    for entry in data:
        score = entry['_source']['Review']['ReviewRating']
        if score in histo:
            histo[score] += 1
        else:
            histo[score] = 1

    if normalize == True:
        for key, value in histo.items():
            histo[key] = 1.0 * value / len(data)
    print(histo)
    return histo


def computeTextStatistics(data, superior_threshold=None, inferior_threshold=None):
    histo_char = {}
    histo_word = {}
    histo_category = {}

    sup_threshold = 0
    inf_threshold = 0

    for entry in data:
        text = entry['_source']['Review']['ReviewBody']
        category = entry['_source']['Product']['ProductCategory']
        chars = len(text)
        words = len(text.split(" "))


        if superior_threshold != None and words > superior_threshold:
            sup_threshold += 1

        if inferior_threshold != None and words < inferior_threshold:
            inf_threshold += 1
        
        if chars in histo_char:
            histo_char[chars] += 1
        else:
            histo_char[chars] = 1
    
        if words in histo_word:
            histo_word[words] += 1
        else:
            histo_word[words] = 1

        if category in histo_category:
            histo_category[category] += 1
        else:
            histo_category[category] = 1

    return histo_char, histo_word, histo_category, sup_threshold, inf_threshold


def computeDatasetStatistics(data, superior_threshold=None, inferior_threshold=None):

    histo_scores = computeScoreHistogram(data)
    histo_chars, histo_words, histo_category, sup_threshold, inf_threshold = computeTextStatistics(data, superior_threshold, inferior_threshold)

    print("Reviews with number of words over", superior_threshold, "=", sup_threshold, "percentage =", 100.0*sup_threshold/len(data))
    print("Reviews with number of words under", inferior_threshold, "=", inf_threshold, "percentage =", 100.0*inf_threshold/len(data))
    print(histo_category)
    plt.bar(histo_scores.keys(), histo_scores.values(), 1.0, color='g')
    plt.title("Scores")    
    plt.show()

    plt.bar(histo_chars.keys(), histo_chars.values(), 1.0, color='g')        
    plt.title("Chars")
    plt.show()


    plt.bar(histo_words.keys(), histo_words.values(), 1.0, color='g')        
    plt.title("Words")
    plt.show()

# split the dataset in 5 vs ALL -> 1,2,3,4 -> label 0
#                                        5 -> label 1
# input -> dataset list of jsons
# output -> dataset list of jsons
def splitData5vAll(data):
    new_data = copy.deepcopy(data)
    for entry in new_data:
        if entry['_source']['Review']['ReviewRating'] == 5:
            entry['_source']['Review']['ReviewRating'] = 1
        else:
            entry['_source']['Review']['ReviewRating'] = 0
    return new_data


# save the dataset
# input -> dataset list of jsons, filename to save
def saveData(data, filename):
    
    with open(filename, 'w') as outfile:
        for entry in data:
            json.dump(entry, outfile)
            outfile.write("\n")
    
# get features from data
# input -> data list of json
# sample_majority -> sample or not from majority class
# sample_count -> how many entries to sample from majority class
# set seed -> random seed value
# output -> list of dicts | one entry is a dict with features and labels
def getFeatures(data, use_review_text=True, sample_majority=False, sample_count=0, seed=None, majority_class=3):
    
    if sample_majority == False:
        train_list = []
        for data_entry in data:
            train_entry = {}
            if use_review_text == True:
                train_entry['features:review_text'] = data_entry['_source']['Review']['ReviewBody']
            train_entry['label'] = data_entry['_source']['Review']['ReviewRating']
            train_list.append(train_entry)
        return train_list
    
    elif sample_majority == True:

        majority_list = []
        for data_entry in data:
            majority_entry = {}
            if data_entry['_source']['Review']['ReviewRating'] == majority_class:
                if use_review_text == True:
                    majority_entry['features:review_text'] = data_entry['_source']['Review']['ReviewBody']
                majority_entry['label'] = data_entry['_source']['Review']['ReviewRating']
                majority_list.append(majority_entry)

        random.seed(seed)
        sampled_majority_list = random.sample(majority_list, sample_count)
        random.seed()

        train_list = []
        for data_entry in data:
            train_entry = {}
            if data_entry['_source']['Review']['ReviewRating'] != majority_class:
                if use_review_text == True:
                    train_entry['features:review_text'] = data_entry['_source']['Review']['ReviewBody']
                train_entry['label'] = data_entry['_source']['Review']['ReviewRating']
                # train_list.append(train_entry)
                sampled_majority_list.append(train_entry)
        
        # train_list.extend(sampled_majority_list)
        train_list = sampled_majority_list
        return train_list

# get processed features and labels
# input -> features
# output -> list of processed features, list of labels, dict of class_weights
def processFeatures(data, bert_proc):
    features = []
    labels = []

    iids = []
    sids = []
    i = 0
    for entry in data:
        review_text = entry["features:review_text"]
        input_ids, segment_ids = bert_proc.process_text(review_text)
        iids.append(input_ids)
        sids.append(segment_ids)
        labels.append(entry['label'])    
    
    features = [np.array(iids), np.array(sids)]
    class_weights = class_weight.compute_class_weight('balanced', np.unique(labels), labels)
    class_weights = class_weights.astype(np.float32)
    return features, labels, class_weights

# get processed features and labels from texst
# input -> features
# output -> list of processed features, list of labels, dict of class_weights
def processFeaturesRawText(data, bert_proc):
    features = []
    iids = []
    sids = []
    i = 0
    for entry in data:
        review_text = entry
        input_ids, segment_ids = bert_proc.process_text(review_text)
        iids.append(input_ids)
        sids.append(segment_ids)
    
    features = [np.array(iids), np.array(sids)]
    return features

# split data in train dev test split using stratified 
# input -> data
# output -> train, dev, test data
def splitTrainDevTest(data):
    
    train_data = []
    dev_data = []
    test_data = []

    full_indices = np.array(range(len(data)))
    full_classes = np.array(list(map(lambda x: x['_source']['Review']['ReviewRating'], data)))

    sss = StratifiedShuffleSplit(n_splits=1, test_size=0.1)

    for tr, te in sss.split(full_indices, full_classes):
        aux_train_indexes = tr
        test_indexes = te

    aux_train_data = []    
    for i in test_indexes:
        test_data.append(data[i])
    for i in aux_train_indexes:
        aux_train_data.append(data[i])


    indices = np.array(range(len(aux_train_data)))
    classes = np.array(list(map(lambda x: x['_source']['Review']['ReviewRating'], aux_train_data)))

    sss_ = StratifiedShuffleSplit(n_splits=1, test_size=0.111111)

    for tr, de in sss_.split(indices, classes):
        train_indexes = tr
        dev_indexes = de

    for i in dev_indexes:
        dev_data.append(aux_train_data[i])
    for i in train_indexes:
        train_data.append(aux_train_data[i])


    print(len(train_data), len(dev_data), len(test_data), len(train_data) + len(dev_data) + len(test_data), len(data))
    print(len(list(set(train_indexes) & set(dev_indexes) & set(test_indexes))))

    return train_data, dev_data, test_data

# split the dataset in 4 classes -> 1 -> label 0
#                                   2,3 -> label 1
#                                   4 -> label 2
#                                   5 -> label 3
# input -> dataset list of jsons
# output -> dataset list of jsons
def splitData4Classes(data):
    new_data = copy.deepcopy(data)
    for entry in new_data:
        if entry['_source']['Review']['ReviewRating'] == 1:
            entry['_source']['Review']['ReviewRating'] = 0
        elif entry['_source']['Review']['ReviewRating'] == 2 or entry['_source']['Review']['ReviewRating'] == 3:
            entry['_source']['Review']['ReviewRating'] = 1
        elif entry['_source']['Review']['ReviewRating'] == 4:
            entry['_source']['Review']['ReviewRating'] = 2
        elif entry['_source']['Review']['ReviewRating'] == 5:
            entry['_source']['Review']['ReviewRating'] = 3
    return new_data


class FScoreCallback(Callback):
    def __init__(self, dataset, steps, labels):

        super().__init__()
        self.steps = steps

        self.dataset = dataset
        self.labels_int = []
        for x in labels:
            self.labels_int.append(np.argmax(x))

    def on_test_end(self, epoch, logs={}):
        y_pred = []
        y_true = self.labels_int
        predict_results = self.model.predict(self.dataset, steps=self.steps)
        for prediction in predict_results:
            y_pred.append(np.argmax(prediction))

        print()
        print(classification_report(y_true, y_pred, digits=4))


def compute_parameters(model_folder_path):

    # define input
    input_ids = tf.keras.layers.Input(shape=(64), dtype=tf.int32, name="input_ids")
    segment_ids = tf.keras.layers.Input(shape=(64), dtype=tf.int32, name="segment_ids")

    import BertModel
    import tensorflow.keras as keras
    import bert

    # define model    
    bert_model = BertModel.BertModel(model_folder_path, 64)
    bert_output = bert_model.bert_layer([input_ids, segment_ids])
    cls_output = keras.layers.Lambda(lambda seq: seq[:, 0, :])(bert_output)
    cls_drop = keras.layers.Dropout(0.1)(cls_output)
    fc1 = keras.layers.Dense(units=100, activation="relu")(cls_drop)
    prediction = keras.layers.Dense(units=10, activation="softmax")(fc1)

    # build model
    model = keras.Model(inputs=[input_ids, segment_ids], outputs=prediction)
    model.build(input_shape=[(None, 64), (None, 64)])
    # load pretrained
    bert.load_bert_weights(bert_model.bert_layer, model_folder_path+"bert_model.ckpt")

    model.compile(optimizer=keras.optimizers.Adam(lr=0.1), loss = 'categorical_crossentropy', metrics = [tf.keras.metrics.categorical_accuracy])
    model.summary()

    from tensorflow.python.keras.utils.layer_utils import count_params
    trainable_count = count_params(model.trainable_weights)
    non_trainable_count = count_params(model.non_trainable_weights)

    print(trainable_count/1e6)
    print(non_trainable_count)

    # return model, bert_model


def build_reallife_corpus(model_folder_path):

    new_model_folder_path = "/".join(model_folder_path.split("/")[:-2])
    new_model_folder_path = os.path.join(new_model_folder_path, "reallife")
    
    train_data = readJson(model_folder_path+"train.json")
    train_data = clean_dict(train_data)
    new_train_data = add_last_sentence_to_data(train_data)
    new_train_data_over = perform_oversampling(new_train_data)
    print(len(train_data), len(new_train_data), len(new_train_data_over))
    saveData(new_train_data_over, os.path.join(new_model_folder_path, "train.json"))


    dev_data = readJson(model_folder_path+"dev.json")
    dev_data = clean_dict(dev_data)
    new_dev_data = add_last_sentence_to_data(dev_data)
    new_dev_data_over = perform_oversampling(new_dev_data)
    print(len(dev_data), len(new_dev_data), len(new_dev_data_over))
    saveData(new_dev_data_over, os.path.join(new_model_folder_path, "dev.json"))

    test_data = readJson(model_folder_path+"test.json")
    test_data = clean_dict(test_data)
    new_test_data = add_last_sentence_to_data(test_data)
    new_test_data_over = perform_oversampling(new_test_data)
    print(len(test_data), len(new_test_data), len(new_test_data_over))
    saveData(new_test_data_over, os.path.join(new_model_folder_path, "test.json"))
    

def add_last_sentence_to_data(data):
    new_data = copy.deepcopy(data)
    new_entries = []
    count = 0
    for entry in new_data:
        review_text = entry['_source']['Review']['ReviewBody']
        sentences = sent_tokenize(review_text)
        if len(sentences) > 1:
            # add new entry to dataset
            new_entry = copy.deepcopy(entry)
            new_entry['_source']['Review']['ReviewBody'] = sentences[-1]
            new_entry['_score'] = 2
            new_entries.append(new_entry)
            if entry == new_entry:
                print(entry)
                print(new_entry)
                sys.exit()
            count += 1
    # print(new_entries)
    new_data.extend(new_entries)
    return new_data


def perform_oversampling(data):
    new_data = copy.deepcopy(data)
    new_entries = []
    counter = [0,0,0,0,0]
    for entry in new_data:
        label = entry['_source']['Review']['ReviewRating']
        counter[label-1] += 1
    
    while True:

        random_entry = random.choice(data)
        random_label = random_entry['_source']['Review']['ReviewRating']

        if counter[random_label-1] == counter[-1]:
            continue
        
        else:
            new_entries.append(random_entry)
            counter[random_label-1] += 1
            
            if counter[0] == counter[1] and counter[1] == counter[2] and counter[2] == counter[3] and counter[3] == counter[4]:
                break

    print(counter)
    new_data.extend(new_entries)
    return new_data


def clean_dict(data):
    new_data = copy.deepcopy(data)
    for entry in new_data:
        del entry["_index"]
        del entry["_type"]
        del entry["_id"]
        del entry["_score"]
        del entry["_source"]["Review"]["ReviewTitle"]
        del entry["_source"]["Review"]["ReviewDate"]
        del entry["_source"]["Review"]["ReviewProductVerified"]
        del entry["_source"]["Product"]
    return new_data



if __name__ == "__main__":

    # data = readJson("../Dataset/Reviews/4Classes/train.json")
    # computeDatasetStatistics(data, 32, 32)

    # print("--------------------------DEV--------------------------")
    # data = readJson("../Dataset/Reviews/4Classes/dev.json")
    # computeDatasetStatistics(data, 32, 32)

    # print("--------------------------TEST--------------------------")
    # data = readJson("../Dataset/Reviews/4Classes/test.json")
    # computeDatasetStatistics(data, 32, 32)

    # compute_parameters("../Models/raw/small/clean/trained_512/ro2/")
    # sys.exit()

    # # split data
    # raw = readJson("../Dataset/Reviews/all_reviews.json", original=True)
    # # computeDatasetStatistics(raw, 256, 256)
    # train_data, dev_data, test_data = splitTrainDevTest(raw)
    # saveData(train_data, "../Dataset/Reviews/emag_train.json")
    # saveData(dev_data, "../Dataset/Reviews/emag_dev.json")
    # saveData(test_data, "../Dataset/Reviews/emag_test.json")


    # raw = readJson("../Dataset/Reviews/all_reviews.json", original=True)
    train_data = readJson("../Dataset/Reviews/emag_train.json")
    # computeDatasetStatistics(train_data, 256, 256)
    dev_data = readJson("../Dataset/Reviews/emag_dev.json")
    test_data = readJson("../Dataset/Reviews/emag_test.json")
    
    computeScoreHistogram(train_data, normalize=True)
    split_train = splitData4Classes(train_data)
    computeScoreHistogram(split_train, normalize=True)
    saveData(split_train, "../Dataset/Reviews/4Classes/train.json")

    computeScoreHistogram(dev_data, normalize=True)
    split_dev = splitData4Classes(dev_data)
    computeScoreHistogram(split_dev, normalize=True)
    saveData(split_dev, "../Dataset/Reviews/4Classes/dev.json")

    computeScoreHistogram(test_data, normalize=True)
    split_test = splitData4Classes(test_data)
    computeScoreHistogram(split_test, normalize=True)
    saveData(split_test, "../Dataset/Reviews/4Classes/test.json")


    


import os
from rb.similarity.word_vector_model import WordVectorModel

from rb.core.lang import Lang
from rb.similarity.vector_model import VectorModelType


class LDA(WordVectorModel):

    def __init__(self, name: str, lang: Lang, dim: int = 300, check_updates=True):
        WordVectorModel.__init__(self, VectorModelType.LDA, name, lang, dim, check_updates=check_updates)  

    


import heapq
import math
import os
import numpy as np
from collections import defaultdict
from functools import reduce
from sklearn.metrics.pairwise import cosine_similarity
from typing import Dict, Iterable, List, Union

from rb.utils.rblogger import Logger
from rb.core.lang import Lang
from rb.parser.spacy_parser import SpacyParser
from rb.core.text_element import TextElement
from rb.core.sentence import Sentence
from rb.core.document import Document
from rb.similarity.vector_model_factory import get_default_model
from rb.similarity.vector_model import VectorModel
import itertools
from sklearn.cluster import KMeans

logger = Logger.get_logger()
logger.propagate = False


def process(sentences: List[Sentence], model: VectorModel, k_ratio: float = 0.2):
    """
    :param tokenized_sentences: every sentence is represented as a list of tokens (strings or word embeddings)
    :param k_ratio: represents the ratio between the number of clusters and the number of docs
    :return: tuple(similarity matrix, dictionary of clusters)
    """
    n = len(sentences)
    similarity_matrix = np.zeros([n, n])
    for i, j in itertools.combinations(range(n), 2):
        sim = model.similarity(sentences[i], sentences[j])
        similarity_matrix[i, j] = similarity_matrix[j, i] = sim
    
    assert (0 < k_ratio <= 1), "A value between 0 and 1 must be provided for k_ratio"
    no_clusters = math.ceil(k_ratio * n)
    km = KMeans(n_clusters=no_clusters, max_iter=20).fit(similarity_matrix)
    clusters = km.labels_

    """
    Note: 'clusters' is a list of size n which has values from 0 to k-1 and maps every
    sentence index to the belonging cluster index
    e.g. clusters=[1 1 0 1 1 0 1 0 1] for k=2
    """
    clusters_dict = {}
    for i, cluster in enumerate(clusters):
        if cluster not in clusters_dict:
            clusters_dict[cluster] = []
        clusters_dict[cluster].append(i)
    return similarity_matrix, clusters_dict

def coverage_function(similarity_matrix, summary_indices: List, alpha_constant: int = 5) -> float:
    coverage_result = 0.0
    if len(summary_indices) == 0:
        return coverage_result
    
    n = len(similarity_matrix)
    alpha = 1.0 * alpha_constant / n
    for i in range(n):
        summary_coverage = sum([similarity_matrix[i][j] for j in summary_indices])
        corpus_coverage = sum(similarity_matrix[i])
        coverage_result += min(summary_coverage, alpha * corpus_coverage)
    """
    Note: The explanation behind min formula is that the same value is returned
    when i becomes saturated or is sufficiently covered by summary.
    """
    return coverage_result

def diversity_reward_function(similarity_matrix, clusters: Dict, summary_indices: List) -> float:
    diversity_reward = 0.0
    if len(summary_indices) == 0:
        return diversity_reward

    no_clusters = len(clusters)
    for k in range(no_clusters):
        common_indices = list(set(summary_indices) & set(clusters[k]))
        if len(common_indices) == 0:
            continue
        cluster_reward = sum(max(sum(similarity_matrix[j]) / len(similarity_matrix[j]), 0)  for j in common_indices)
        diversity_reward += math.sqrt(cluster_reward)
    
    return diversity_reward
    
def greedy_submodular_maximization(similarity_matrix, clusters, costs_dict, summary_size, alpha, tradeoff_coefficient, word_count=None):
    n = len(similarity_matrix)
    prev_score = 0

    G = set()
    U = list(range(n))
    
    while len(U) > 0:
        temp_scores = []
        for l in U:
            # c_l = costs_dict[l]
            temp_summary_indices = list(G) + [l]

            coverage_result = coverage_function(similarity_matrix, temp_summary_indices, alpha)
            diversity_reward = diversity_reward_function(similarity_matrix, clusters, temp_summary_indices)
            current_score = coverage_result + tradeoff_coefficient * diversity_reward

            win = (current_score - prev_score)# / pow(c_l, 0.3)
            temp_scores += [(win, l, current_score)]

        (best_win, k, score) = max(temp_scores, key=lambda t: t[0])
        # if sum([costs_dict[i] for i in G]) + costs_dict[k] <= word_count:
        if len(G) + 1 <= summary_size:
            if best_win >= 0:
                G.add(k)
                prev_score = score
        else:
            break
        U = [i for i in U if i != k]
    return sorted(G)

def summarize(doc: Union[str, TextElement], lang: Lang=None, ratio=0.2, word_count=None, tradeoff_coefficient=5, k_ratio=0.1, alpha=3) -> List[Sentence]:
    if isinstance(doc, str):
        assert lang is not None, "lang parameter is required for str docs"
        doc = Document(lang, doc)
    else:
        lang = doc.lang
    doc: TextElement
    model = get_default_model(lang)
    sentences = doc.get_sentences()
    summary_size = int(len(sentences) * ratio) if word_count is None else 1
    assert 0 < summary_size < len(sentences)

    # get_doc_index = dict(zip(range(len(doc_indices)), doc_indices))
    similarity_matrix, clusters = process(sentences, model, k_ratio=k_ratio)

    # costs_ = {i: costs_dict[get_doc_index[i]] for i in range(tokenized_sentences_len)}
    summary_indices = greedy_submodular_maximization(similarity_matrix, clusters, None, summary_size, alpha, tradeoff_coefficient, word_count)

    return [sentences[idx] for idx in summary_indices]




from rb.processings.ro_corrections.get_exceptions import *

def get_pronoun(tag):
	return tag[2], tag[4], tag[3]

def get_verb(tag):
	if len(tag) < 6:
		number = ""
	else:
		number = tag[5]
	return tag[4], number

def get_noun(tag):
	#not enough tag details
	if tag != "Np":
		return '3', tag[3], tag[2]

def get_adjective(tag):
	return "-", tag[4], tag[3]

def get_numeral(tag):
	return "-", tag[3], tag[2]

def get_unstated_article(tag):
	return "-", tag[3], tag[2]

def get_person_number_gender(spn):
	person = ""
	number = ""
	gender = ""
	if spn[0] == "3":
		person = "a 3-a"
	else:
		person = spn[0]
	if spn[1] == "s":
		number = "singular"
	else:
		number = "plural"
	if spn[2] == "f":
		gender = "feminin"
	else:
		gender = "masculin"
	return person, number, gender


#word could be a noun or a predicate
def get_multiple_subject(word, relation, exceptions):
	subjects = []
	if relation == "conj":
		subjects = [word]
	for child in word.children:
		if child.dep_ == relation:
			subjects.append(child)

	persons = are_describing_persons(subjects, exceptions)
	if len(persons) != 0:
		p = get_person_for_multiple_subject(persons)
		if all_feminine_gender(persons):
			return p, "p", "f"
		else :
			return p, "p", "m"

	all_nouns = True
	for subject in subjects:
		if subject.tag_[:2] != "Nc" and subject.tag_[:2] != "Np":
			all_nouns = False
			break
	if all_nouns:
		#if there is at least one subject at gender masculin and plural -> near noun is the influent, else will be feminin plural
		if at_least_one_plural_masculin_noun(subjects):
			near_subject = subjects[len(subjects) - 1]
			(person, number, gender) = get_noun(near_subject.tag_)
			return "3", "p", gender
		if at_least_one_feminin_noun(subjects):
			return "3", "p", "f"
		return "3", "p", "m"


def all_feminine_gender(persons):
	for person in persons:
		if person.tag_[:2] == "Nc" or person.tag_[:2] == "Np":
			nounOrPronoun = get_noun(person.tag_)
		elif person.tag_[0] == "P":
			nounOrPronoun = get_pronoun(person.tag_)
		if nounOrPronoun[2] == "m" or nounOrPronoun[2] == "-":
			return False
	return True

def get_person_for_multiple_subject(persons):
	first = False
	second = False
	for person in persons:
		if person.tag_[:2] == "Nc" or person.tag_[:2] == "Np":
			nounOrPronoun = get_noun(person.tag_)
		elif person.tag_[0] == "P":
			nounOrPronoun = get_pronoun(person.tag_)
		if nounOrPronoun[0] == "1":
			first = True
			break
		if nounOrPronoun[0] == "2":
			second = True
	if first:
		return "1"
	elif second: 
		return "2"
	else:
		return "3"


def at_least_one_plural_masculin_noun(subjects):
	for subject in subjects:
		(person, number, gender) = get_noun(subject.tag_)
		if number == "p" and gender == "m":
			return True

	return False

def at_least_one_feminin_noun(subjects):
	for subject in subjects:
		(person, number, gender) = get_noun(subject.tag_)
		if gender == "f":
			return True

	return False

#al that are in conjunction relation
def get_multiple_predicates(root):
	predicates = [root]
	for child in root.children:
		if child.dep_ == "conj" and has_own_subject(child) == False:
			predicates.append(child)

	return predicates

#because 2 sentences that are in coordinative relation will have verbs in conjunction, even if they have own subject
def has_own_subject(predicate):
	for child in predicate.children:
		if child.dep_ == "nsubj":
			return True
	return False



from rb.core.lang import Lang
from rb.core.document import Document
from rb.complexity.complexity_index import ComplexityIndex, compute_indices
from rb.complexity.cohesion.adj_cohesion import AdjCohesion
from rb.similarity.word2vec import Word2Vec
from rb.cna.cna_graph import CnaGraph
from typing import Tuple, List
from sklearn.svm import SVR
import pickle
import os
import csv
from rb.utils.rblogger import Logger
from typing import List, Tuple
import numpy as np

logger = Logger.get_logger()

class Fluctuations:

    def __init__(self):
        pass
    


from rb.processings.pipeline.estimator import Classifier, Regressor
from rb.processings.pipeline.dataset import Dataset, Task
from typing import List, Dict
from sklearn import neural_network

class MLPClassifier(Classifier):
    def __init__(self, dataset: Dataset, tasks: List[Task], params: Dict[str, str]):
        super().__init__(dataset, tasks, params)
        self.model = neural_network.MLPClassifier(hidden_layer_sizes=params["hidden_layer_sizes"])
        self.hidden = params["hidden_layer_sizes"]
        
    @classmethod
    def parameters(cls):
        return {
            "hidden_layer_sizes": [32, 64, 96, 128]
        }
    
    def __str__(self):
        return f"MLP - {self.hidden}"

class MLPRegressor(Regressor):
    def __init__(self, dataset: Dataset, tasks: List[Task], params: Dict[str, str]):
        super().__init__(dataset, tasks, params)
        self.model = neural_network.MLPRegressor(hidden_layer_sizes=params["hidden_layer_sizes"])
        self.hidden = params["hidden_layer_sizes"]
        
    @classmethod
    def parameters(cls):
        return {
            "hidden_layer_sizes": [32, 64, 96, 128]
        }
    
    def __str__(self):
        return f"MLP - {self.hidden}"


import numpy as np
import random
import math
import time
from collections import Counter


class StringKernel():

    def __init__(self):
        pass
    @staticmethod
    def compute_kernel_two_strings(string1, string2, ngram_range_min, ngram_range_max, clusters=None):
        pass

    @staticmethod
    def compute_kernel_string_listofstrings(string1, strings, ngram_range_min, ngram_range_max, clusters=None):
        pass

    @staticmethod
    def compute_kernel_listofstrings(strings, ngram_range_min, ngram_range_max, normalize=False):
        pass

    @staticmethod
    def run(dataset, ngram_range_min, ngram_range_max, normalize=False, clusters=None):
        pass

class IntersectionStringKernel(StringKernel):

    def __init__(self):
        super().__init__()

    @staticmethod
    def compute_kernel_two_strings(string1, string2, ngram_range_min, ngram_range_max, clusters=None):

        if clusters != None and len(clusters) > 0:
            return IntersectionStringKernel.compute_kernel_two_strings_clusters2(string1, string2, ngram_range_min, ngram_range_max, clusters)

        ngrams = {}
        for char_index, char in enumerate(string1):
            for d in range(ngram_range_min, ngram_range_max+1):
                if char_index + d <= len(string1):
                    ngram = string1[char_index:char_index+d]
                    if ngram not in ngrams:
                        ngrams[ngram] = 1
                    else:
                        ngrams[ngram] = ngrams[ngram] + 1
        kernel = 0
        for char_index, char in enumerate(string2):
            for d in range(ngram_range_min, ngram_range_max+1):
                if char_index + d <= len(string2):
                    ngram = string2[char_index:char_index+d]
                    if (ngram in ngrams) and (ngrams[ngram] > 0):
                        kernel += 1
                        ngrams[ngram] = ngrams[ngram] - 1
        return 1.0*kernel

    @staticmethod
    def compute_kernel_two_strings_clusters(string1, string2, ngram_range_min, ngram_range_max, clusters):

        ngrams = {}
        ngrams_in_clusters = {}

        for char_index, char in enumerate(string1):
            for d in range(ngram_range_min, ngram_range_max + 1):
                if char_index + d <= len(string1):
                    ngram = string1[char_index:char_index + d]
                    if ngram not in ngrams_in_clusters:
                        found = False
                        for cluster in clusters:
                            if ngram in cluster:
                                found = True
                                break
                        ngrams_in_clusters[ngram] = found
                    if ngrams_in_clusters[ngram] == False:
                        ngrams_in_clusters[ngram] = False
                        if ngram not in ngrams:
                            ngrams[ngram] = 1
                        else:
                            ngrams[ngram] = ngrams[ngram] + 1


        kernel = 0
        for char_index, char in enumerate(string2):
            for d in range(ngram_range_min, ngram_range_max + 1):
                if char_index + d <= len(string2):
                    ngram = string2[char_index:char_index + d]
                    if (ngram in ngrams) and (ngrams[ngram] > 0):
                        kernel += 1
                        ngrams[ngram] = ngrams[ngram] - 1

        d1 = {}
        d2 = {}

        for cl, _ in enumerate(clusters):
            d1[cl] = 0
            d2[cl] = 0

        for cl_index, cl in enumerate(clusters):
            for ngram in cl:
                if len(ngram) >= ngram_range_min and len(ngram) <= ngram_range_max:
                    d1[cl_index] = d1[cl_index] + string1.count(ngram)
                    d2[cl_index] = d2[cl_index] + string2.count(ngram)

        for cl, _ in enumerate(clusters):
            kernel += min(d1[cl],d2[cl])
        return kernel

    @staticmethod
    def compute_kernel_two_strings_clusters2(string1, string2, ngram_range_min, ngram_range_max, clusters):

        index = 0
        clusters_dict = {}
        for cluster in clusters:
            for ngram in cluster:
                clusters_dict[ngram] = index
            index += 1

        s1_ngram = Counter()
        s2_ngram = Counter()

        for char_index, char in enumerate(string1):
            for d in range(ngram_range_min, ngram_range_max + 1):
                if char_index + d <= len(string1):
                    ngram = string1[char_index:char_index + d]
                    if ngram not in clusters_dict:
                        clusters_dict[ngram] = index
                        index += 1
                    s1_ngram[clusters_dict[ngram]] += 1

        for char_index, char in enumerate(string2):
            for d in range(ngram_range_min, ngram_range_max + 1):
                if char_index + d <= len(string2):
                    ngram = string2[char_index:char_index + d]
                    if ngram not in clusters_dict:
                        clusters_dict[ngram] = index
                        index += 1
                    s2_ngram[clusters_dict[ngram]] += 1

        kernel = 0
        for c in s1_ngram:
            kernel += min(s1_ngram[c], s2_ngram[c])

        return kernel

    @staticmethod
    def compute_kernel_string_listofstrings(string1, strings, ngram_range_min, ngram_range_max, normalize=False, clusters=None):
        kernels = []
        for string2 in strings:
            pr = IntersectionStringKernel.compute_kernel_two_strings(string1, string2, ngram_range_min, ngram_range_max, clusters)
            if normalize == True:
                i = IntersectionStringKernel.compute_kernel_two_strings(string1, string1, ngram_range_min, ngram_range_max, clusters)
                j = IntersectionStringKernel.compute_kernel_two_strings(string2, string2, ngram_range_min, ngram_range_max, clusters)
                if i == 0.0 or j == 0.0:
                    kernels.append(0.0)
                else:
                    kernels.append(pr / math.sqrt(i*j))
            elif normalize == False:
                kernels.append(pr)

        kernels = np.array(kernels)
        return kernels

    @staticmethod
    def compute_kernel_listofstrings(strings, ngram_range_min, ngram_range_max, normalize=False):
        kernels = []
        for string1 in strings:
            kernels.append(IntersectionStringKernel.compute_kernel_string_listofstrings(string1, strings, ngram_range_min, ngram_range_max))

        if normalize == True:
            for i, aux in enumerate(kernels):
                for j, _ in enumerate(aux):
                    print (i,j, kernels[i][j], kernels[i][i], kernels[j][j], math.sqrt(kernels[i][i] * kernels[j][j]))
                    kernels[i][j] = kernels[i][j] /  math.sqrt(kernels[i][i] * kernels[j][j])
                    print (kernels[i][j])

        kernels = np.array(kernels)
        return kernels

    @staticmethod
    def run(dataset, ngram_range_min, ngram_range_max, normalize=False, clusters=None):
        correct = 0
        total = 0

        for entry_index, entry in enumerate(dataset.data):
            if len(entry.answer_pool) == 0:
                total += 1
                continue
            question = entry.question
            similarity_scores = IntersectionStringKernel.compute_kernel_string_listofstrings(question, entry.answer_pool, ngram_range_min, ngram_range_max, normalize, clusters)
            max_indexes = np.argwhere(similarity_scores == np.max(similarity_scores)).flatten().tolist()
            random_max_index = random.choice(max_indexes)
            if entry.answer_pool[random_max_index] in entry.correct_answer:
                correct += 1
            total += 1

        # print ("Intersection kernel (", ngram_range_min, ngram_range_max, normalize, ") =", 1.0* correct / total, flush=True)
        return 1.0 * correct / total

class PresenceStringKernel(StringKernel):

    def __init__(self):
        super().__init__()

    @staticmethod
    def compute_kernel_two_strings(string1, string2, ngram_range_min, ngram_range_max, clusters=None):
        if clusters != None and len(clusters) > 0:
            return PresenceStringKernel.compute_kernel_two_strings_clusters2(string1, string2, ngram_range_min, ngram_range_max, clusters)

        ngrams = {}
        for char_index, char in enumerate(string1):
            for d in range(ngram_range_min, ngram_range_max+1):
                if char_index + d <= len(string1):
                    ngram = string1[char_index:char_index+d]
                    ngrams[ngram] = 1
        kernel = 0
        for char_index, char in enumerate(string2):
            for d in range(ngram_range_min, ngram_range_max+1):
                if char_index + d <= len(string2):
                    ngram = string2[char_index:char_index+d]
                    if (ngram in ngrams):
                        kernel += 1
                        ngrams.pop(ngram)
        return 1.0*kernel

    @staticmethod
    def compute_kernel_two_strings_clusters(string1, string2, ngram_range_min, ngram_range_max, clusters):

        ngrams = {}
        ngrams_in_clusters = {}

        for char_index, char in enumerate(string1):
            for d in range(ngram_range_min, ngram_range_max + 1):
                if char_index + d <= len(string1):
                    ngram = string1[char_index:char_index + d]
                    if ngram not in ngrams_in_clusters:
                        found = False
                        for cluster in clusters:
                            if ngram in cluster:
                                found = True
                                break
                        ngrams_in_clusters[ngram] = found
                    if ngrams_in_clusters[ngram] == False:
                        if ngram not in ngrams:
                            ngrams[ngram] = 1
                        else:
                            ngrams[ngram] = ngrams[ngram] + 1

        kernel = 0
        for char_index, char in enumerate(string2):
            for d in range(ngram_range_min, ngram_range_max + 1):
                if char_index + d <= len(string2):
                    ngram = string2[char_index:char_index + d]
                    if (ngram in ngrams):
                        kernel += 1
                        ngrams.pop(ngram)

        d1 = {}
        d2 = {}

        for cl, _ in enumerate(clusters):
            d1[cl] = 0
            d2[cl] = 0

        for cl_index, cl in enumerate(clusters):
            for ngram in cl:
                if len(ngram) >= ngram_range_min and len(ngram) <= ngram_range_max:
                    d1[cl_index] = d1[cl_index] + string1.count(ngram)
                    d2[cl_index] = d2[cl_index] + string2.count(ngram)

        for cl, _ in enumerate(clusters):
            if d1[cl] * d2[cl] == 0:
                continue
            kernel += 1
        return kernel

    @staticmethod
    def compute_kernel_two_strings_clusters2(string1, string2, ngram_range_min, ngram_range_max, clusters):

        index = 0
        clusters_dict = {}
        for cluster in clusters:
            for ngram in cluster:
                clusters_dict[ngram] = index
            index += 1

        s1_ngram = Counter()
        s2_ngram = Counter()

        for char_index, char in enumerate(string1):
            for d in range(ngram_range_min, ngram_range_max + 1):
                if char_index + d <= len(string1):
                    ngram = string1[char_index:char_index + d]
                    if ngram not in clusters_dict:
                        clusters_dict[ngram] = index
                        index += 1
                    s1_ngram[clusters_dict[ngram]] += 1

        for char_index, char in enumerate(string2):
            for d in range(ngram_range_min, ngram_range_max + 1):
                if char_index + d <= len(string2):
                    ngram = string2[char_index:char_index + d]
                    if ngram not in clusters_dict:
                        clusters_dict[ngram] = index
                        index += 1
                    s2_ngram[clusters_dict[ngram]] += 1

        kernel = 0
        for c in s1_ngram:
            if s1_ngram[c] * s2_ngram[c] >= 1:
                kernel += 1
        return kernel

    @staticmethod
    def compute_kernel_string_listofstrings(string1, strings, ngram_range_min, ngram_range_max, normalize=False, clusters=None):
        kernels = []
        for string2 in strings:
            pr = PresenceStringKernel.compute_kernel_two_strings(string1, string2, ngram_range_min, ngram_range_max, clusters)
            if normalize == True:
                i = PresenceStringKernel.compute_kernel_two_strings(string1, string1, ngram_range_min, ngram_range_max, clusters)
                j = PresenceStringKernel.compute_kernel_two_strings(string2, string2, ngram_range_min, ngram_range_max, clusters)
                if i == 0.0 or j == 0.0:
                    kernels.append(0.0)
                else:
                    kernels.append(pr / math.sqrt(i*j))
            elif normalize == False:
                kernels.append(pr)

        kernels = np.array(kernels)
        return kernels

    @staticmethod
    def compute_kernel_listofstrings(strings, ngram_range_min, ngram_range_max, normalize=False):
        kernels = []
        for string1 in strings:
            kernels.append(PresenceStringKernel.compute_kernel_string_listofstrings(string1, strings, ngram_range_min, ngram_range_max))

        if normalize == True:
            for i, aux in enumerate(kernels):
                for j, _ in enumerate(aux):
                    print (i,j, kernels[i][j], kernels[i][i], kernels[j][j], math.sqrt(kernels[i][i] * kernels[j][j]))
                    kernels[i][j] = kernels[i][j] /  math.sqrt(kernels[i][i] * kernels[j][j])
                    print (kernels[i][j])

        kernels = np.array(kernels)
        return kernels

    @staticmethod
    def run(dataset, ngram_range_min, ngram_range_max, normalize=False, clusters=None):
        correct = 0
        total = 0

        for entry_index, entry in enumerate(dataset.data):
            if len(entry.answer_pool) == 0:
                total += 1
                continue
            question = entry.question
            similarity_scores = PresenceStringKernel.compute_kernel_string_listofstrings(question, entry.answer_pool, ngram_range_min, ngram_range_max, normalize, clusters)
            max_indexes = np.argwhere(similarity_scores == np.max(similarity_scores)).flatten().tolist()
            random_max_index = random.choice(max_indexes)
            if entry.answer_pool[random_max_index] in entry.correct_answer:
                correct += 1
            total += 1

        # print("Presence kernel (", ngram_range_min, ngram_range_max, normalize, ") =", 1.0 * correct / total, flush=True)
        return 1.0* correct / total

class SpectrumStringKernel(StringKernel):

    def __init__(self):
        super().__init__()

    @staticmethod
    def compute_kernel_two_strings(string1, string2, ngram_range_min, ngram_range_max, clusters=None):

        if clusters != None and len(clusters) > 0:
            return SpectrumStringKernel.compute_kernel_two_strings_clusters2(string1, string2, ngram_range_min, ngram_range_max, clusters)

        ngrams = {}
        for char_index, char in enumerate(string1):
            for d in range(ngram_range_min, ngram_range_max+1):
                if char_index + d <= len(string1):
                    ngram = string1[char_index:char_index+d]
                    if ngram not in ngrams:
                        ngrams[ngram] = 1
                    else:
                        ngrams[ngram] = ngrams[ngram] + 1
        kernel = 0
        for char_index, char in enumerate(string2):
            for d in range(ngram_range_min, ngram_range_max+1):
                if char_index + d <= len(string2):
                    ngram = string2[char_index:char_index+d]
                    if (ngram in ngrams):
                        kernel += ngrams[ngram]

        return 1.0*kernel

    @staticmethod
    def compute_kernel_two_strings_clusters(string1, string2, ngram_range_min, ngram_range_max, clusters):

        ngrams = {}
        ngrams_in_clusters = {}

        for char_index, char in enumerate(string1):
            for d in range(ngram_range_min, ngram_range_max+1):
                if char_index + d <= len(string1):
                    ngram = string1[char_index:char_index+d]
                    if ngram not in ngrams_in_clusters:
                        found = False
                        for cluster in clusters:
                            if ngram in cluster:
                                found = True
                                break
                        ngrams_in_clusters[ngram] = found
                    if ngrams_in_clusters[ngram] == False:
                        ngrams_in_clusters[ngram] = False
                        if ngram not in ngrams:
                            ngrams[ngram] = 1
                        else:
                            ngrams[ngram] = ngrams[ngram] + 1

        kernel = 0
        for char_index, char in enumerate(string2):
            for d in range(ngram_range_min, ngram_range_max+1):
                if char_index + d <= len(string2):
                    ngram = string2[char_index:char_index+d]
                    if (ngram in ngrams):
                        kernel += ngrams[ngram]

        d1 = {}
        d2 = {}

        for cl, _ in enumerate(clusters):
            d1[cl] = 0
            d2[cl] = 0

        for cl_index, cl in enumerate(clusters):
            for ngram in cl:
                if len(ngram) >= ngram_range_min and len(ngram) <= ngram_range_max:
                    d1[cl_index] = d1[cl_index] + string1.count(ngram)
                    d2[cl_index] = d2[cl_index] + string2.count(ngram)

        for cl, _ in enumerate(clusters):
            kernel += d1[cl] * d2[cl]
        return kernel

    @staticmethod
    def compute_kernel_two_strings_clusters2(string1, string2, ngram_range_min, ngram_range_max, clusters):

        index = 0
        clusters_dict = {}
        for cluster in clusters:
            for ngram in cluster:
                clusters_dict[ngram] = index
            index += 1

        s1_ngram = Counter()
        s2_ngram = Counter()

        for char_index, char in enumerate(string1):
            for d in range(ngram_range_min, ngram_range_max + 1):
                if char_index + d <= len(string1):
                    ngram = string1[char_index:char_index + d]
                    if ngram not in clusters_dict:
                        clusters_dict[ngram] = index
                        index += 1
                    s1_ngram[clusters_dict[ngram]] += 1

        for char_index, char in enumerate(string2):
            for d in range(ngram_range_min, ngram_range_max + 1):
                if char_index + d <= len(string2):
                    ngram = string2[char_index:char_index + d]
                    if ngram not in clusters_dict:
                        clusters_dict[ngram] = index
                        index += 1
                    s2_ngram[clusters_dict[ngram]] += 1

        kernel = 0
        for c in s1_ngram:
            # print (c, s1_ngram[c], s2_ngram[c])
            kernel += s1_ngram[c] * s2_ngram[c]
        return kernel

    @staticmethod
    def compute_kernel_string_listofstrings(string1, strings, ngram_range_min, ngram_range_max, normalize=False, clusters=None):
        kernels = []
        for string2 in strings:
            pr = SpectrumStringKernel.compute_kernel_two_strings(string1, string2, ngram_range_min, ngram_range_max, clusters)
            if normalize == True:
                i = SpectrumStringKernel.compute_kernel_two_strings(string1, string1, ngram_range_min, ngram_range_max, clusters)
                j = SpectrumStringKernel.compute_kernel_two_strings(string2, string2, ngram_range_min, ngram_range_max, clusters)
                if i == 0.0 or j == 0.0:
                    kernels.append(0.0)
                else:
                    kernels.append(pr / math.sqrt(i*j))
            elif normalize == False:
                kernels.append(pr)

        kernels = np.array(kernels)
        return kernels

    @staticmethod
    def compute_kernel_listofstrings(strings, ngram_range_min, ngram_range_max, normalize=False):
        kernels = []
        for string1 in strings:
            kernels.append(SpectrumStringKernel.compute_kernel_string_listofstrings(string1, strings, ngram_range_min, ngram_range_max))

        if normalize == True:
            for i, aux in enumerate(kernels):
                for j, _ in enumerate(aux):
                    print (i,j, kernels[i][j], kernels[i][i], kernels[j][j], math.sqrt(kernels[i][i] * kernels[j][j]))
                    kernels[i][j] = kernels[i][j] /  math.sqrt(kernels[i][i] * kernels[j][j])
                    print (kernels[i][j])

        kernels = np.array(kernels)
        return kernels

    @staticmethod
    def run(dataset, ngram_range_min, ngram_range_max, normalize=False, clusters=None):
        correct = 0
        total = 0

        for entry_index, entry in enumerate(dataset.data):
            if len(entry.answer_pool) == 0:
                total += 1
                continue
            question = entry.question
            similarity_scores = SpectrumStringKernel.compute_kernel_string_listofstrings(question, entry.answer_pool, ngram_range_min, ngram_range_max, normalize, clusters)
            max_indexes = np.argwhere(similarity_scores == np.max(similarity_scores)).flatten().tolist()
            random_max_index = random.choice(max_indexes)
            if entry.answer_pool[random_max_index] in entry.correct_answer:
                correct += 1
            total += 1

        # print("Spectrum kernel (", ngram_range_min, ngram_range_max, normalize, ") =", 1.0 * correct / total, flush=True)
        return 1.0 * correct / total


if __name__ == "__main__":

    a = "walking kin lorem ipsum dem dolort adsr su u  how are you to come to me to go to stay some things we want to change but no change vauce forte nefaster pentru a recladi nimeni nu ma munca"
    b = ["de vreo miscare de rezinta hufhre stii cum un singur patid o singura germanie wtf is this i am wathcing ot stay stalin wtf is do you want to thisn"]

    t = time.time()
    for _ in range(100):
        c = IntersectionStringKernel.compute_kernel_string_listofstrings(a, b, 1, 10, normalize=False)
    print (c, time.time()-t)

    t = time.time()
    for _ in range(100):
        c = IntersectionStringKernel.compute_kernel_string_listofstrings(a, b, 1, 10, normalize=False, clusters=[["aaaa"],["bbbb"]])
    print(c , time.time() - t)




# Copyright 2017 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""In-Graph Beam Search Implementation.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import namedtuple
import numpy as np

import tensorflow as tf
from tensorflow.python.util import nest  # pylint: disable=E0611


class BeamSearchState(
    namedtuple("BeamSearchState", ["log_probs", "finished", "lengths"])):
  """State for a single step of beam search.

  Args:
    log_probs: The current log probabilities of all beams
    finished: A boolean vector that specifies which beams are finished
    lengths: Lengths of all beams
  """
  pass


class BeamSearchStepOutput(
    namedtuple("BeamSearchStepOutput",
               ["scores", "predicted_ids", "beam_parent_ids"])):
  """Outputs for a single step of beam search.

  Args:
    scores: Score for each beam, a float32 vector
    predicted_ids: predictions for this step step, an int32 vector
    beam_parent_ids: an int32 vector containing the beam indices of the
      continued beams from the previous step
  """
  pass


class BeamSearchConfig(
    namedtuple("BeamSearchConfig", [
        "beam_width", "vocab_size", "eos_token", "length_penalty_weight",
        "choose_successors_fn"
    ])):
  """Configuration object for beam search.

  Args:
    beam_width: Number of beams to use, an integer
    vocab_size: Output vocabulary size
    eos_token: The id of the EOS token, used to mark beams as "done"
    length_penalty_weight: Weight for the length penalty factor. 0.0 disables
      the penalty.
    choose_successors_fn: A function used to choose beam successors based
      on their scores. Maps from (scores, config) => (chosen scores, chosen_ids)
  """
  pass


def gather_tree_py(values, parents):
  """Gathers path through a tree backwards from the leave nodes. Used
  to reconstruct beams given their parents."""
  beam_length = values.shape[0]
  num_beams = values.shape[1]
  res = np.zeros_like(values)
  res[-1, :] = values[-1, :]
  for beam_id in range(num_beams):
    parent = parents[-1][beam_id]
    for level in reversed(range(beam_length - 1)):
      res[level, beam_id] = values[level][parent]
      parent = parents[level][parent]
  return np.array(res).astype(values.dtype)


def gather_tree(values, parents):
  """Tensor version of gather_tree_py"""
  res = tf.compat.v1.py_func(
      func=gather_tree_py, inp=[values, parents], Tout=values.dtype)
  res.set_shape(values.get_shape().as_list())
  return res


def create_initial_beam_state(config):
  """Creates an instance of `BeamState` that can be used on the first
  call to `beam_step`.

  Args:
    config: A BeamSearchConfig

  Returns:
    An instance of `BeamState`.
  """
  return BeamSearchState(
      log_probs=tf.zeros([config.beam_width]),
      finished=tf.zeros(
          [config.beam_width], dtype=tf.bool),
      lengths=tf.zeros(
          [config.beam_width], dtype=tf.int32))


def length_penalty(sequence_lengths, penalty_factor):
  """Calculates the length penalty according to
  https://arxiv.org/abs/1609.08144

   Args:
    sequence_lengths: The sequence length of all hypotheses, a tensor
      of shape [beam_size, vocab_size].
    penalty_factor: A scalar that weights the length penalty.

  Returns:
    The length penalty factor, a tensor fo shape [beam_size].
   """
  return tf.truncatediv((5. + tf.cast(sequence_lengths, dtype=tf.float32))**penalty_factor, (5. + 1.)
                **penalty_factor)


def hyp_score(log_probs, sequence_lengths, config):
  """Calculates scores for beam search hypotheses.
  """

  # Calculate the length penality
  length_penality_ = length_penalty(
      sequence_lengths=sequence_lengths,
      penalty_factor=config.length_penalty_weight)
  score = log_probs / length_penality_
  return score


def choose_top_k(scores_flat, config):
  """Chooses the top-k beams as successors.
  """
  next_beam_scores, word_indices = tf.nn.top_k(scores_flat, k=config.beam_width)
  return next_beam_scores, word_indices


def nest_map(inputs, map_fn, name=None):
  """Applies a function to (possibly nested) tuple of tensors.
  """
  if nest.is_sequence(inputs):
    inputs_flat = nest.flatten(inputs)
    y_flat = [map_fn(_) for _ in inputs_flat]
    outputs = nest.pack_sequence_as(inputs, y_flat)
  else:
    outputs = map_fn(inputs)
  if name:
    outputs = tf.identity(outputs, name=name)
  return outputs


def mask_probs(probs, eos_token, finished):
  """Masks log probabilities such that finished beams
  allocate all probability mass to eos. Unfinished beams remain unchanged.

  Args:
    probs: Log probabiltiies of shape `[beam_width, vocab_size]`
    eos_token: An int32 id corresponding to the EOS token to allocate
      probability to
    finished: A boolean tensor of shape `[beam_width]` that specifies which
      elements in the beam are finished already.

  Returns:
    A tensor of shape `[beam_width, vocab_size]`, where unfinished beams
    stay unchanged and finished beams are replaced with a tensor that has all
    probability on the EOS token.
  """
  vocab_size = tf.shape(input=probs)[1]
  finished_mask = tf.expand_dims(tf.cast(1. - tf.cast(finished, dtype=tf.float32), dtype=tf.float32), 1)
  # These examples are not finished and we leave them
  non_finished_examples = finished_mask * probs
  # All finished examples are replaced with a vector that has all
  # probability on EOS
  finished_row = tf.one_hot(
      eos_token,
      vocab_size,
      dtype=tf.float32,
      on_value=0.,
      off_value=tf.float32.min)
  finished_examples = (1. - finished_mask) * finished_row
  return finished_examples + non_finished_examples


def beam_search_step(time_, logits, beam_state, config):
  """Performs a single step of Beam Search Decoding.

  Args:
    time_: Beam search time step, should start at 0. At time 0 we assume
      that all beams are equal and consider only the first beam for
      continuations.
    logits: Logits at the current time step. A tensor of shape `[B, vocab_size]`
    beam_state: Current state of the beam search. An instance of `BeamState`
    config: An instance of `BeamSearchConfig`

  Returns:
    A new beam state.
  """

  # Calculate the current lengths of the predictions
  prediction_lengths = beam_state.lengths
  previously_finished = beam_state.finished

  # Calculate the total log probs for the new hypotheses
  # Final Shape: [beam_width, vocab_size]
  probs = tf.nn.log_softmax(logits)
  probs = mask_probs(probs, config.eos_token, previously_finished)
  total_probs = tf.expand_dims(beam_state.log_probs, 1) + probs

  # Calculate the continuation lengths
  # We add 1 to all continuations that are not EOS and were not
  # finished previously
  lengths_to_add = tf.one_hot([config.eos_token] * config.beam_width,
                              config.vocab_size, 0, 1)
  add_mask = (1 - tf.cast(previously_finished, dtype=tf.int32))
  
  lengths_to_add = tf.expand_dims(add_mask, 1) * lengths_to_add
  new_prediction_lengths = tf.expand_dims(prediction_lengths,
                                          1) + lengths_to_add

  # Calculate the scores for each beam
  scores = hyp_score(
      log_probs=total_probs,
      sequence_lengths=new_prediction_lengths,
      config=config)

  scores_flat = tf.reshape(scores, [-1])
  # During the first time step we only consider the initial beam
  scores_flat = tf.cond(
      pred=tf.convert_to_tensor(value=time_) > 0, true_fn=lambda: scores_flat, false_fn=lambda: scores[0])

  # Pick the next beams according to the specified successors function
  next_beam_scores, word_indices = config.choose_successors_fn(scores_flat,
                                                               config)
  next_beam_scores.set_shape([config.beam_width])
  word_indices.set_shape([config.beam_width])

  # Pick out the probs, beam_ids, and states according to the chosen predictions
  total_probs_flat = tf.reshape(total_probs, [-1], name="total_probs_flat")
  next_beam_probs = tf.gather(total_probs_flat, word_indices)
  next_beam_probs.set_shape([config.beam_width])
  next_word_ids = tf.math.mod(word_indices, config.vocab_size)
  next_beam_ids = tf.truncatediv(word_indices, config.vocab_size)

  # Append new ids to current predictions
  next_finished = tf.logical_or(
      tf.gather(beam_state.finished, next_beam_ids),
      tf.equal(next_word_ids, config.eos_token))

  # Calculate the length of the next predictions.
  # 1. Finished beams remain unchanged
  # 2. Beams that are now finished (EOS predicted) remain unchanged
  # 3. Beams that are not yet finished have their length increased by 1
  lengths_to_add = tf.cast(tf.not_equal(next_word_ids, config.eos_token), dtype=tf.int32)
  lengths_to_add = (1 - tf.cast(next_finished, dtype=tf.int32)) * lengths_to_add
  next_prediction_len = tf.gather(beam_state.lengths, next_beam_ids)
  next_prediction_len += lengths_to_add

  next_state = BeamSearchState(
      log_probs=next_beam_probs,
      lengths=next_prediction_len,
      finished=next_finished)

  output = BeamSearchStepOutput(
      scores=next_beam_scores,
      predicted_ids=next_word_ids,
      beam_parent_ids=next_beam_ids)

  return output, next_state



from rb.comprehension.utils.graph.cm_node_do import CmNodeDO, CmNodeType
from rb.comprehension.utils.memory.word_activation import WordActivation
from rb.comprehension.utils.graph.cm_graph_do import CmGraphDO
from rb.comprehension.utils.graph.graph_metrics import GraphMetrics

import numpy as np

from typing import Dict

ActivationMap = Dict[CmNodeDO, float]


class HistoryKeeper:

    def __init__(self):
        self.activation_history = []
        self.graph_metrics_history = []
        self.unique_word_list = []

    def save_state(self, activation_map: ActivationMap, graph: CmGraphDO):
        self.save_scores(activation_map)
        self.graph_metrics_history.append(GraphMetrics(graph))

    def save_scores(self, activation_map: ActivationMap) -> None:
        activation_auxiliary = {}

        for key in activation_map.keys():
            activation_auxiliary[key.word] = WordActivation(activation_map[key], key.active)

        self.activation_history.append(activation_auxiliary)

    def save_nodes(self, graph: CmGraphDO) -> None:
        for node in graph.node_list:
            self.add_node_if_not_existing(node)

    def add_node_if_not_existing(self, node: CmNodeDO) -> None:
        exists = False
        for index, current_node in enumerate(self.unique_word_list):
            if current_node == node:
                if current_node.node_type != CmNodeType.TextBased and node.node_type == CmNodeType.TextBased:
                    self.unique_word_list[index] = node
                exists = True
                break

        if not exists:
            self.unique_word_list.append(node)

    def compute_statistics(self):
        mean_activation_for_active_nodes = self.compute_mean_activation_for_active_nodes()
        mean_activation_for_all_nodes = self.compute_mean_activation_for_all_nodes()
        mean_degree_centrality = self.compute_mean_degree_centrality()
        mean_closeness_centrality = self.compute_mean_closeness_centrality()
        mean_betweenness_centrality = self.compute_mean_betweenness_centrality()
        mean_harmonic_centrality = self.compute_mean_harmonic_centrality()
        mean_active_nodes_percentage = self.compute_mean_active_nodes_percentage()
        mean_density = self.compute_mean_density()
        mean_modularity = self.compute_mean_modularity()

        return [mean_activation_for_active_nodes, mean_activation_for_all_nodes, mean_degree_centrality,
                mean_closeness_centrality, mean_betweenness_centrality, mean_harmonic_centrality,
                mean_active_nodes_percentage, mean_density, mean_modularity]

    def compute_mean_activation_for_active_nodes(self):
        total_activation_sum = 0.0
        total_count = len(self.activation_history)
        for activation_dictionary in self.activation_history:
            total_activation_sum += np.mean([word_activation.activation_value
                                             for word_activation in activation_dictionary.values()
                                             if word_activation.active])

        return total_activation_sum / total_count

    def compute_mean_activation_for_all_nodes(self):
        total_activation_sum = 0.0
        total_count = len(self.activation_history)
        for activation_dictionary in self.activation_history:
            total_activation_sum += np.mean([word_activation.activation_value
                                             for word_activation in activation_dictionary.values()])

        return total_activation_sum / total_count

    def compute_mean_centrality(self, centrality_function):
        total_centrality = 0.0
        total_count = len(self.graph_metrics_history)
        for graph_metrics in self.graph_metrics_history:
            method = getattr(graph_metrics, centrality_function, None)
            centrality_dictionary = method()
            total_centrality += np.mean(list(centrality_dictionary.values()))

        return total_centrality / total_count

    def compute_mean_degree_centrality(self):
        return self.compute_mean_centrality("degree_centrality")

    def compute_mean_closeness_centrality(self):
        return self.compute_mean_centrality("closeness_centrality")

    def compute_mean_betweenness_centrality(self):
        return self.compute_mean_centrality("betweenness_centrality")

    def compute_mean_harmonic_centrality(self):
        return self.compute_mean_centrality("harmonic_centrality")

    def compute_mean_number_of_active_nodes(self):
        return np.mean([len(gm.G.nodes) for gm in self.graph_metrics_history])

    def compute_mean_number_of_active_edges(self):
        return np.mean([len(gm.G.edges) for gm in self.graph_metrics_history])

    def compute_mean_active_nodes_percentage(self):
        return np.mean([len(graph_metrics.G.nodes) / (len(self.activation_history[index]))
                        for index, graph_metrics in enumerate(self.graph_metrics_history)])

    def compute_mean_metrics(self, metric_function):
        return np.mean([getattr(graph_metrics, metric_function, None)()
                        for graph_metrics in self.graph_metrics_history])

    def compute_mean_density(self):
        return self.compute_mean_metrics("density")

    def compute_mean_modularity(self):
        return self.compute_mean_metrics("modularity")




from enum import Enum, auto

class NamedEntityWikiEnum(Enum):
    PER = auto()
    LOC = auto()
    ORG = auto()
    MISC = auto()


class NamedEntityONEnum(Enum):
    PERSON = auto()
    NORP = auto()
    FAC	= auto()
    ORG	= auto()
    GPE	= auto()
    LOC	= auto()
    PRODUCT	= auto()
    EVENT = auto()
    WORK_OF_ART	= auto()
    LAW	= auto()
    LANGUAGE = auto()
    DATE = auto()
    TIME = auto()
    PERCENT	= auto()
    MONEY = auto()
    QUANTITY = auto()
    ORDINAL	= auto()
    CARDINAL = auto()


from typing import List, Dict
from datetime import datetime, timedelta

from rb.core.lang import Lang
from rb.core.cscl.community import Community
from rb.cna.cna_graph import CnaGraph
from rb.processings.cscl.participant_evaluation import perform_sna, evaluate_interaction, evaluate_involvement, evaluate_textual_complexity

from rb.core.cscl.cscl_criteria import CsclCriteria

from rb.utils.rblogger import Logger
from rb.processings.cscl.community_processing import compute_new_threads_metrics

SECONDS_IN_DAY = 60 * 60 * 24
SECONDS_IN_MONTH = SECONDS_IN_DAY * 30

def determine_subcommunities(community: Community, start_date: datetime = None, end_date: datetime = None):
    if start_date is None:
        start_date = community.first_contribution_date
    if start_date is None:
        return
    if end_date is None:
        end_date = community.last_contribution_date
    if end_date is None:
        return
    week = 1
    while start_date < end_date:
        next_date = min(start_date + timedelta(days=7), end_date)
        subcommunity = extract_subcommunity(
            parent_community=community,
            subcommunity_start=start_date,
            subcommunity_end=next_date)
        print(f"week {week}: {len(subcommunity.components)} conversations")
        week += 1
        start_date = next_date
        community.add_subcommunity(subcommunity)
        
def extract_subcommunity(parent_community: Community, subcommunity_start: datetime, subcommunity_end: datetime) -> Community:
    subcommunity = Community(lang=parent_community.lang,
                             container=parent_community,
                             community=parent_community.get_community_raw(),
                             start_date=subcommunity_start,
                             end_date=subcommunity_end)
    subcommunity.graph = CnaGraph(docs=[subcommunity], models=parent_community.graph.models)
    # evaluate_interaction(subcommunity)
    evaluate_involvement(subcommunity)
    # evaluate_textual_complexity(subcommunity)
    perform_sna(subcommunity)
    # compute_new_threads_metrics(subcommunity)
    return subcommunity



#pylint: disable=import-error
import os
import pickle
import sys

import absl
import numpy as np
import rb.processings.sentiment.utils_new as utils
import tensorflow as tf
import tensorflow.keras as keras
from rb.core.lang import Lang
from rb.processings.encoders.bert import BertWrapper
from rb.processings.sentiment.BertRegression import BertRegression
from rb.utils.downloader import check_version, download_model
from tensorflow.keras.models import load_model
from transformers import AutoTokenizer, TFAutoModelForSequenceClassification


class SentimentAnalysis(object):
	"""
	Wrapper for Sentiment Analysis
    """

	def __init__(self, lang: Lang, model_type="base", max_seq_len=128, check_updates = True):
		# load model
		self.lang = lang
		self.max_seq_len = min(max_seq_len, 512)
		if self.lang is Lang.RO:
			self._load_model(model_type, check_updates)
		else:
			self.tokenizer = AutoTokenizer.from_pretrained("nlptown/bert-base-multilingual-uncased-sentiment") 
			self.model = TFAutoModelForSequenceClassification.from_pretrained("nlptown/bert-base-multilingual-uncased-sentiment", from_pt=True)
	
	# loads best sentiment model
	def _load_model(self, model_type, check_updates = True):
		self.bert_wrapper = BertWrapper(self.lang, max_seq_len=self.max_seq_len, model_name=model_type, check_updates=check_updates, custom_model=True)
		if check_updates and check_version(Lang.RO, ["models", "sentiment", model_type]):
			download_model(Lang.RO, ["models", "sentiment", model_type])
		model_path = f"resources/{self.lang.value}/models/sentiment/{model_type}/"
		self.model = BertRegression(bert_wrapper=self.bert_wrapper, bert_trainable=False, bert_pooling_type="cls",
					learning_rate=0, hidden_size=[128,64], restore_model="",
					optimizer="adam", loss="mse", dropout_rate=0.1, models_path=model_path).model
		
	def process_text(self, text):

		if isinstance(text, str):
			text = [text]

		if self.lang is Lang.RO:
			features = utils.processFeaturesRawText(text, self.bert_wrapper)
			predictions = self.model.predict(features)
			return [x[0] / 5.0 for x in predictions]
		else:
			features = self.tokenizer(text)
			features = {key: np.array(value) for key, value in features.items()}
			predictions = tf.nn.softmax(self.model(features).logits).numpy()
			return [sum(i / 5 * p for i, p in enumerate(pred)) for pred in predictions]
		


import tensorflow as tf
from rb.processings.gec.transformer.utils import positional_encoding
from rb.processings.gec.transformer.decoder_layer import DecoderLayer

class Decoder(tf.keras.layers.Layer):
    def __init__(self, num_layers, d_model, num_heads, dff, target_vocab_size,
               maximum_position_encoding, rate=0.1):
        super(Decoder, self).__init__()

        self.d_model = d_model
        self.num_layers = num_layers
        
        self.embedding = tf.keras.layers.Embedding(target_vocab_size, d_model)
        self.pos_encoding = positional_encoding(maximum_position_encoding, d_model)
        
        self.dec_layers = [DecoderLayer(d_model, num_heads, dff, rate) 
                        for _ in range(num_layers)]
        self.dropout = tf.keras.layers.Dropout(rate)
        
    def call(self, x, enc_output, training, 
           look_ahead_mask, padding_mask):

        seq_len = tf.shape(x)[1]
        attention_weights = {}
        
        x = self.embedding(x)  # (batch_size, target_seq_len, d_model)
        x *= tf.math.sqrt(tf.cast(self.d_model, tf.float32))
        x += self.pos_encoding[:, :seq_len, :]
        
        x = self.dropout(x, training=training)

        for i in range(self.num_layers):
            x, block1, block2 = self.dec_layers[i](x, enc_output, training,
                                                look_ahead_mask, padding_mask)
        
            attention_weights['decoder_layer{}_block1'.format(i+1)] = block1
            attention_weights['decoder_layer{}_block2'.format(i+1)] = block2
        
        # x.shape == (batch_size, target_seq_len, d_model)
        return x, attention_weights


import csv
import math
import os
from heapq import heapify, heappop
from typing import Dict, Iterable, List, Tuple

import numpy as np
from joblib import Parallel, delayed
from rb.cna.cna_graph import CnaGraph
from rb.complexity.complexity_index import ComplexityIndex, compute_indices
from rb.core.document import Document
from rb.core.lang import Lang
from rb.core.meta_document import MetaDocument
from rb.processings.pipeline.bert_classifier import BertClassifier
from rb.processings.pipeline.dataset import Dataset, TargetType, Task
from rb.processings.pipeline.estimator import Estimator
from rb.processings.pipeline.mlp import MLPClassifier, MLPRegressor
from rb.processings.pipeline.random_forrest import RandomForest
from rb.processings.pipeline.ridge_regression import RidgeRegression
from rb.processings.pipeline.svm import SVM
from rb.processings.pipeline.svr import SVR
from rb.similarity.vector_model import VectorModel
from rb.similarity.vector_model_factory import get_default_model
from rb.utils.rblogger import Logger
from scipy.stats import f_oneway, pearsonr

CLASSIFIERS = [SVM, RandomForest, MLPClassifier]
REGRESSORS = [SVR, RidgeRegression, MLPRegressor]
    
logger = Logger.get_logger()
        
 
def construct_document(lang: Lang, text: str) -> MetaDocument:
    sections = ["\n".join(line for line in section.split("\n") if len(line) > 50) for section in text.split("\n\n")]
    return MetaDocument(lang, sections)
    
    
def compute_features(doc: MetaDocument):
    model = get_default_model(doc.lang)
    cna_graph = CnaGraph(docs=doc, models=[model])
    for section in doc.components:
        compute_indices(doc=section, cna_graph=cna_graph)     
    return {
        feature: np.mean([section.indices[feature] for section in doc.components]) 
        for feature in doc.components[0].indices.keys()
    }
        
def construct_documents(dataset: List[str], lang: Lang) -> List[Dict[ComplexityIndex, float]]:
    logger.info("Constructing documents..")
    docs = Parallel(n_jobs=1)( \
        delayed(construct_document)(lang, text) \
        for text in dataset)
    get_default_model(lang)
    logger.info("Constructing graphs..")
    return Parallel(n_jobs=-1, prefer="processes")( \
        delayed(compute_features)(doc) \
        for doc in docs)

    
def filter_rare(dataset: Dataset):
    features = []
    for index in dataset.features:
        values = [indices[index] for indices in dataset.train_features]
        zeros = sum(1 for val in values if val == 0)
        if zeros / len(values) < 0.2:
            features.append(index)
    dataset.features = features

def preprocess(folder: str, targets_file: str, lang: Lang, limit: int = None) -> Dataset:
    files = {filename.replace(".txt", "").strip(): open(os.path.join(folder, filename), "rt", encoding='utf-8', errors='ignore').read().strip() 
             for filename in os.listdir(folder)
             if not filename.startswith(".")}
    names = []
    texts = []
    targets = []
    with open(targets_file, "rt", encoding='utf-8') as f:
        reader = csv.reader(f, delimiter=',')
        header = next(reader)
        for line in reader:
            filename = line[0].replace(".txt", "").strip()
            if filename not in files:
                print(filename)
                continue
            names.append(filename)
            texts.append(files[filename])
            targets.append(line[1:])
            if limit is not None and len(names) == limit:
                break
    dataset = Dataset(lang, names, texts, targets)
    
    dataset.all_features = construct_documents(dataset.texts, lang)
    dataset.features = list(dataset.all_features[0].keys())
    dataset.split(0.2)
    filter_rare(dataset)
    logger.info("Removing colinear..")
    for task in dataset.tasks:
        remove_colinear(dataset, task)
    return dataset
    # dataset.save_features("features.csv")

def correlation_with_targets(feature: ComplexityIndex, dataset: Dataset, task: Task) -> float:
    values = [indices[feature] for indices in dataset.train_features]
    if task.type is TargetType.FLOAT:
        corr, p = pearsonr(values, task.train_values)
        return abs(corr)
    values_per_class = {}
    for val, target in zip(values, task.train_values):
        if target not in values_per_class:
            values_per_class[target] = []
        values_per_class[target].append(val)
    return f_oneway(*values_per_class.values())

def remove_colinear(dataset: Dataset, task: Task) -> None:
    heap = []
    for i, a in enumerate(dataset.features[:-1]):
        for j, b in enumerate(dataset.features[i+1:]):
            values_a = [indices[a] for indices in dataset.train_features]
            values_b = [indices[b] for indices in dataset.train_features]
            corr, p = pearsonr(values_a, values_b)
            if math.isnan(corr):
                continue
            heap.append((-corr, i, j))
    heapify(heap)
    
    correlations = [correlation_with_targets(feature, dataset, task) 
                    for feature in dataset.features]
    task.mask = [True] * len(dataset.features)
    while len(heap) > 0:
        inv_corr, i, j = heappop(heap)
        if not task.mask[i] or not task.mask[j]:
            continue
        if inv_corr < -0.9:
            if correlations[i] > correlations[j]:
                task.mask[j] = False
            else:
                task.mask[i] = False


def next_config(estimator, parameters: Dict[str, List]) -> Iterable[Dict[str, str]]:
    solution  = [-1] * len(parameters)
    keys = list(parameters.keys())
    current = 0
    while current >= 0:
        if current == len(parameters):
            config = {keys[i]: parameters[keys[i]][j] for i, j in enumerate(solution)}
            if estimator.valid_config(config):
                yield config
            current -= 1
        if current < 0:
            break
        if solution[current] == len(parameters[keys[current]]) - 1:
            solution[current] = -1
            current -= 1
            continue
        solution[current] += 1
        current += 1
            

def grid_search(dataset: Dataset, task: Task) -> Estimator:
    if task.type is TargetType.FLOAT:
        estimators = REGRESSORS
        default_best = (None, float('inf'))
        better = lambda a, b: a < b
    else:
        estimators = CLASSIFIERS
        default_best = (None, 0)
        better = lambda a, b: a > b
    results = []    
    for estimator in estimators:
        parameters = estimator.parameters()
        best = default_best
        for config in next_config(estimator, parameters):
            model = estimator(dataset, [task], config)
            acc = model.cross_validation()
            if better(acc, best[1]):
                best = (model, acc)
            print(f"{model} - {acc}")
        model = best[0]
        score = model.evaluate()
        results.append((model, score))
    return results

def bert_grid_search(dataset: Dataset, use_indices=True, use_mask=True, shared=True) -> Tuple[Estimator, List[float]]:
    parameters = BertClassifier.parameters()
    best = (None, 0, float('inf'))
    for config in next_config(BertClassifier, parameters):
        model = BertClassifier(dataset, dataset.tasks, config, use_indices, use_mask, shared)
        epoch, loss = model.cross_validation()
        print(f"{config}: loss={loss}")
        if loss < best[2]:
            best = (model, epoch, loss)
    model, epoch, loss = best
    print(f"Best model after {epoch} epochs with loss={loss}")
    model.initialize()
    scores = model.train(epoch)
    return model, scores

def evaluate_bert_models(dataset: Dataset) -> List[Tuple[Estimator, List[float]]]:
    results = []
    results.append(bert_grid_search(dataset, use_indices=False, shared=False))
    results.append(bert_grid_search(dataset, use_indices=True, shared=False, use_mask=False))
    results.append(bert_grid_search(dataset, use_indices=True, use_mask=True, shared=False))
    results.append(bert_grid_search(dataset, use_indices=True, use_mask=False, shared=True))
    results.append(bert_grid_search(dataset, use_indices=True, use_mask=True, shared=True))
    return results



from statistics import mean
from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType   
from typing import List, Callable
from rb.similarity.vector_model import VectorModel
from rb.cna.cna_graph import CnaGraph
from rb.cna.edge_type import EdgeType

from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class InterCohesion(ComplexityIndex):

    """InterCohesion between elements element_type"""
    def __init__(self, lang: Lang, element_type: TextElementType,
            reduce_depth: int, reduce_function: MeasureFunction, cna_graph: CnaGraph):
        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.COHESION,
                                 reduce_depth=reduce_depth, reduce_function=reduce_function,
                                 abbr="InterParCoh")
        self.element_type = element_type
        self.cna_graph = cna_graph    
        if element_type.value > reduce_depth:
            logger.error('For index {} element_type has to be lower or equal than reduce_depth'.format(self))

    def _compute_value(self, element: TextElement) -> float:
        if len(element.components) < 2:
            return True
        sim_values = []
        for i, a in enumerate(element.components[:-1]):
            for b in element.components[(i+1):]:
                sim_edge = self.cna_graph.edges(node=(a, b), edge_type=EdgeType.SEMANTIC, vector_model=None)
                if len(sim_edge) > 0 and len(sim_edge[0]) >= 3:
                    sim_values.append(sim_edge[0][2])
        return mean(sim_values)
    
    def __repr__(self):
        return f"{self.reduce_function_abbr}({self.abbr} / {self.reduce_depth_abbr})"


from enum import Enum

class POS(Enum):
    ADJ = 'ADJ'
    ADP = 'ADP'
    ADV = 'ADV'
    AUX = 'AUX'
    CONJ = 'CONJ'
    CCONJ = 'CCONJ'
    DET = 'DET'
    INTJ = 'INTJ'
    NOUN = 'NOUN'
    NUM = 'NUM'
    PART = 'PART'
    PRON = 'PRON'
    PROPN = 'PROPN'
    PUNCT = 'PUNCT'
    SCONJ = 'SCONJ'
    SYM = 'SYM'
    VERB = 'VERB'
    X = 'X'
    SPACE = 'SPACE'

    def to_wordnet(self) -> str:
        if self is self.ADJ:
            return 'a'
        if self is self.ADV:
            return 'r'
        if self is self.NOUN:
            return 'n'
        if self is self.VERB:
            return 'v'
        return None



from enum import Enum, auto
from rb.core.pos_features.pos_feature import POSFeature
from rb.core.pos_features.ro_pos_features.ro_features_name import RoFeaturesName
from rb.core.lang import Lang
from rb.core.pos import POS
from typing import List
import re


class RoAdpTypeEnum(Enum):
    PREP = auto()


class RoPOSFeatureAdpT(POSFeature):


    _INSTANCE = None

    def __init__(self):
        POSFeature.__init__(self, lang=Lang.RO, name=RoFeaturesName.ADPT,
                            values=None, pos_supported=None)
        self.values = set([ro_adp_type for _, ro_adp_type in RoAdpTypeEnum.__members__.items()])
        self.pos_supported = set([POS.ADP, POS.PUNCT, POS.SYM])

    @classmethod
    def get_instance(cls) -> "RoPOSFeatureAdpT":
        if cls._INSTANCE is None:
            cls._INSTANCE = RoPOSFeatureAdpT()
        return cls._INSTANCE

    def get_values(self, tag: str) -> List[RoAdpTypeEnum]:
        pattern = '(Prep=[a-zA-Z,]*){1}'
        matches = re.findall(pattern, tag)
        if len(matches) == 0: return []
        ro_adp_types = matches[0].split('=')[1].split(',')
        values = []
        for ro_adp_type in ro_adp_types:
            if ro_adp_type == 'Prep':
                values.append(RoAdpTypeEnum.PREP)
        return values


from typing import List
from rb.core.lang import Lang
from rb.core.text_element_type import TextElementType
from rb.complexity.measure_function import MeasureFunction
from rb.cna.cna_graph import CnaGraph


def create(lang: Lang, cna_graph: CnaGraph) -> List["ComplexityIndex"]:
    from rb.complexity.surface.no_words import NoWord
    from rb.complexity.surface.no_unq_words import NoUniqueWordsIndex
    from rb.complexity.surface.no_commas import NoCommas
    from rb.complexity.surface.no_punctuations import NoPunctuations
    from rb.complexity.surface.no_sentences import NoSentences
    from rb.complexity.surface.wd_entropy import WdEntropy
    from rb.complexity.surface.ch_entropy import ChEntropy
    from rb.complexity.surface.ch_ngram_entropy import ChNgramEntropyEnum, ChNgramEntropy

    indices = []

    text_elements = [TextElementType.SENT.value, TextElementType.BLOCK.value, TextElementType.DOC.value]
    measure_functions = [MeasureFunction.AVG, MeasureFunction.STDEV, MeasureFunction.MAX]
    classes = [NoWord, NoUniqueWordsIndex, NoSentences, NoPunctuations, NoCommas, WdEntropy]
    
    for index_class in classes:
        for text_element in text_elements:
            if index_class is NoSentences and text_element == TextElementType.SENT.value:
                continue
            for measure_function in measure_functions:                
                indices.append(index_class(lang, text_element, measure_function))
    
    indices.append(ChNgramEntropy(lang, ChNgramEntropyEnum.TWO, TextElementType.WORD.value, MeasureFunction.AVG))
    indices.append(ChNgramEntropy(lang, ChNgramEntropyEnum.TWO, TextElementType.WORD.value, MeasureFunction.STDEV))
    indices.append(ChNgramEntropy(lang, ChNgramEntropyEnum.TWO, TextElementType.WORD.value, MeasureFunction.MAX))
    
    # if lang is Lang.RO:
    #     #indices.append(NoCacophonies(Lang.RO, TextElementType.SENT.value, MeasureFunction.AVG))
    #     indices.append(NoCommonErrors(Lang.RO, TextElementType.SENT.value, MeasureFunction.AVG))
    
    return indices

    


from rb.core.lang import Lang
from rb.core.document import Document
from rb.complexity.complexity_index import ComplexityIndex, compute_indices
from rb.similarity.word2vec import Word2Vec
from rb.similarity.vector_model import VectorModelType, CorporaEnum, VectorModel
from rb.similarity.vector_model_factory import VECTOR_MODELS, create_vector_model
from typing import Tuple, List
from sklearn.svm import SVR
from collections import Counter
from sklearn import svm
from sklearn.metrics import confusion_matrix, plot_confusion_matrix
import matplotlib.pyplot as plt
import pickle
import os
import csv
import random
from werkzeug import secure_filename
import uuid
from rb.cna.cna_graph import CnaGraph
from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class TextClassifier:


    def __init__(self):
        pass
    
    def get_vector_model(self, lang: Lang = Lang.RO) -> VectorModel:
        global logger
        if lang is Lang.RO:
            vector_model = create_vector_model(Lang.RO, VectorModelType.from_str('word2vec'), "readme")
        elif lang is Lang.EN:
            vector_model = create_vector_model(Lang.EN, VectorModelType.from_str("word2vec"), "coca")
        else:
            logger.error(f'Language {lang.value} is not supported for essay scoring task')
            vector_model = None
        return vector_model
    
    def read_indices(self, base_folder: str = 'categories_readme', lang=Lang.RO) -> List[List[float]]:

        categroies = ['general_stats.csv', 'literature_stats.csv', 'science_stats.csv']
        results = []
        indices = []
        if lang is Lang.RO:
            with open('rb/processings/text_classifier/indices_ro_class.txt', 'rt', encoding='utf-8') as f:
                for line in f:
                    indices.append(line.strip())

        for j, cat in enumerate(categroies):
            essay_r = csv.reader(open(os.path.join(base_folder, cat), 'rt', encoding='utf-8'))
            """ first col is the score """
            for i, row in enumerate(essay_r):
                if i == 0:
                    indices_row = row
                    continue
                
                res = [j]
                for findex in indices:
                    for k, rr in enumerate(row):
                        if indices_row[k].strip() == findex:
                            res.append(rr)
                            break
                results.append(res)
        return results

    def train_svm(self, results: List[List], save_model_file=None):
        total = len(results)
        random.shuffle(results)
        train_samples = int(total * 0.8)

        train = results[:train_samples]
        test = results[train_samples:]

        y = [int(r[0]) for r in train]
        X = [r[1:] for r in train]

        clf = svm.SVC(kernel='poly', degree=14, class_weight={0: 0.1, 1: 0.6, 2: 0.3}).fit(X, y)
        if save_model_file:
            pickle.dump(clf, open(save_model_file, 'wb'))

        dev_out, dev_in = [], []

        for sample_x in test:
            if int(sample_x[0]) == 0 and random.random() < 0.7:
                continue
            dev_out.append(int(sample_x[0]))
            Xx = sample_x[1:]
            dev_in.append(Xx)
        print(Counter(dev_out))
        disp = plot_confusion_matrix(clf, dev_in, dev_out, display_labels=['general', 'science', 'literature', ])
        
        res = clf.predict(dev_in)
        disp.ax_.set_title('Confusion Matrix')
        right, wrong = 0, 0
        for r, clss in zip(res, dev_out):
            if r != clss:
                wrong += 1
            else:
                right += 1            

        logger.info('Acc for classification : {}'.format(right/(wrong + right)))
        plt.show()

    def predict(self, content: str, file_to_svr_model: str, lang=Lang.RO) -> int:

        svr = pickle.load(open(file_to_svr_model, "rb"))

        doc = Document(lang=lang, text=content)
        vector_model = self.get_vector_model(lang=lang)

        cna_graph = CnaGraph(docs=doc, models=[vector_model])
        compute_indices(doc=doc, cna_graph=cna_graph)

        indices = []
        if lang is Lang.RO:
            with open('rb/processings/text_classifier/indices_ro_class.txt', 'rt', encoding='utf-8') as f:
                for line in f:
                    indices.append(line.strip())
        values_indices = []
        for ind in indices: 
            for key, v in doc.indices.items():
                if repr(key) == ind:
                    values_indices.append(v)
                    break

        class_txt = svr.predict([values_indices])[0]
        return class_txt

        


import json
import sys
import matplotlib.pyplot as plt
import copy
import numpy as np
import tensorflow as tf
from sklearn.model_selection import StratifiedShuffleSplit
from sklearn.utils import class_weight
from collections import Counter
import random
from tensorflow.keras.callbacks import Callback
from sklearn.metrics import classification_report
from sklearn.metrics import accuracy_score

# get features from data
# input -> data list of json
# sample_majority -> sample or not from majority class
# sample_count -> how many entries to sample from majority class
# set seed -> random seed value
# output -> list of dicts | one entry is a dict with features and labels
def getFeatures(data, use_review_text=True, sample_majority=False, sample_count=0, seed=None):
    
    if sample_majority == False:
        train_list = []
        for data_entry in data:
            train_entry = {}
            if use_review_text == True:
                train_entry['features:review_text'] = data_entry['_source']['Review']['ReviewBody']
            train_entry['label'] = data_entry['_source']['Review']['ReviewRating']
            train_list.append(train_entry)
        return train_list
    
    elif sample_majority == True:

        majority_list = []
        for data_entry in data:
            majority_entry = {}
            if data_entry['_source']['Review']['ReviewRating'] == 3:
                if use_review_text == True:
                    majority_entry['features:review_text'] = data_entry['_source']['Review']['ReviewBody']
                majority_entry['label'] = data_entry['_source']['Review']['ReviewRating']
                majority_list.append(majority_entry)

        random.seed(seed)
        sampled_majority_list = random.sample(majority_list, sample_count)
        random.seed()

        train_list = []
        for data_entry in data:
            train_entry = {}
            if data_entry['_source']['Review']['ReviewRating'] != 3:
                if use_review_text == True:
                    train_entry['features:review_text'] = data_entry['_source']['Review']['ReviewBody']
                train_entry['label'] = data_entry['_source']['Review']['ReviewRating']
                # train_list.append(train_entry)
                sampled_majority_list.append(train_entry)
        
        # train_list.extend(sampled_majority_list)
        train_list = sampled_majority_list
        return train_list

# get processed features and labels
# input -> features
# output -> list of processed features, list of labels, dict of class_weights
def processFeatures(data, bert_proc):
    features = []

    iids = []
    sids = []
    i = 0
    for entry in data:
        review_text = entry
        input_ids, segment_ids = bert_proc.process_text(review_text)
        iids.append(input_ids)
        sids.append(segment_ids)
    
    features = [np.array(iids), np.array(sids)]
    return features



import tensorflow as tf
import bert
import os

class BertEncoder(tf.keras.layers.Layer):
    def __init__(self, model_dir, d_model, args):
        super(BertEncoder, self).__init__(trainable=False)
        bert_params = bert.params_from_pretrained_ckpt(model_dir)
        self.bert_layer = bert.BertModelLayer.from_params(bert_params, name="bert_layer")
        self.model_dir = model_dir
        tf.compat.v1.logging.info('bert model loaded from {}'.format(model_dir))
        tf.compat.v1.logging.info('bert model params: {}'.format(bert_params))
        # do dummy call to build the model indirectly 
        self.bert_layer([tf.zeros([args.batch_size, args.seq_length], dtype=tf.dtypes.int64),
             tf.zeros([args.batch_size, args.seq_length], dtype=tf.dtypes.int64)])
        bert.load_bert_weights(self.bert_layer, os.path.join(self.model_dir, "bert_model.ckpt"))
        tf.compat.v1.logging.info('bert weights loaded')
        
    def call(self, input_ids, segment_ids, training):
        bert_output = self.bert_layer([input_ids, segment_ids])
        
        return bert_output  # (batch_size, input_seq_len, d_model)


import tensorflow as tf
import os
from os.path import join
from shutil import copyfile
from bert.tokenization.bert_tokenization import FullTokenizer
import tensorflow_datasets as tfds
from typing import Dict, List, Tuple
import numpy as np
from rb.processings.gec.transformer.utils import create_masks
from rb.processings.gec.transformer.serialization import example_encode_text_dataset, get_text_dataset_tf_records
from rb.processings.gec.transformer.serialization import serialize_ids_dataset

args, tokenizer_ro, tokenizer_bert = None, None, None

train_step_signature_np = [tf.TensorSpec(shape=(None, None, None), dtype=tf.int64),
    tf.TensorSpec(shape=(None, None), dtype=tf.int64)]


def construct_flat_datasets(args1, subwords_path):
    global tokenizer_bert, tokenizer_ro, args

    args = args1
    if args.bert:
        tokenizer_bert = FullTokenizer(vocab_file=join(args.bert_model_dir, "vocab.vocab"))
        tokenizer_bert.vocab_size = len(tokenizer_bert.vocab)

    samples = get_text_samples(args)

    if os.path.isfile(subwords_path + '.subwords'): 
        tokenizer_ro  = construct_tokenizer(None, subwords_path, args)
    else:
        tokenizer_ro = construct_tokenizer(list(samples), subwords_path, args)

    sample_train = int(args.total_samples * args.train_dev_split)

    if args.records:
        dataset = tf.data.Dataset.from_generator(generator_tensors_ids_and_segs,
                                        ((tf.int64, tf.int64), tf.int64), 
                                        ((tf.TensorShape([None]), tf.TensorShape([None])), tf.TensorShape([None])))
    else:
        gen_dataset = generator_tensors_ids()
        dataset = list(gen_dataset)
        nr_samples = len(dataset)
        sample_train = int(args.train_dev_split * nr_samples)
        # dataset = tf.convert_to_tensor(dataset, dtype=tf.int64)
        dataset = tf.data.Dataset.from_generator(generator_tensors_ids,
                                        (tf.int64, tf.int64), 
                                        (tf.TensorShape([2, args.seq_length]), tf.TensorShape([args.seq_length])))

    train_dataset = dataset.take(sample_train)
    dev_dataset = dataset.skip(sample_train)

    return train_dataset, dev_dataset

def construct_datasets_gec(args, subwords_path):
    train_dataset, dev_dataset = construct_flat_datasets(args, subwords_path)
    return prepare_datasets(train_dataset, dev_dataset, args)
   
def prepare_datasets(train_dataset, dev_dataset, args):
    train_dataset = train_dataset.shuffle(args.buffer_size).batch(args.batch_size, drop_remainder=True)
    train_dataset = train_dataset.prefetch(tf.data.experimental.AUTOTUNE) # how many batches to prefectch

    dev_dataset = dev_dataset.shuffle(args.buffer_size).batch(args.batch_size, drop_remainder=True)
    return train_dataset, dev_dataset

def construct_tf_records(args1, subwords_path=None):
    """given a txt constructs tf records files + subwords dictionary"""
    global tokenizer_bert, tokenizer_ro
    tf.compat.v1.logging.info('constructing tf records files and vocabularies in {}'.format(args1.tf_records))
    train_dataset, dev_dataset = construct_flat_datasets(args1, subwords_path)

    if not os.path.exists(args1.tf_records):
        os.makedirs(args1.tf_records)
    tokenizer_ro_path = os.path.join(args1.tf_records, 'tokenizer_ro')
    tokenizer_bert_path = os.path.join(args1.tf_records, 'tokenizer_bert.vocab')

    tokenizer_ro.save_to_file(tokenizer_ro_path)
    vocab_file_source = join(args1.bert_model_dir, "vocab.vocab")
    if args1.bert:
        copyfile(vocab_file_source, tokenizer_bert_path)
    
    # todo save bert tokenizer
    serialize_ids_dataset(train_dataset, args1, 'train.tfrecord')
    serialize_ids_dataset(dev_dataset, args1, 'dev.tfrecord')
    tf.compat.v1.logging.info('tf records files and vocabularies constructed in {}'.format(args1.tf_records))

def test_map_numpy(tensor1, tensor2):
    global args
    tensor1 = tensor1.numpy()
    tensor1 *= 2 
    t1 = tf.convert_to_tensor(tensor1, dtype=tf.int64)
    tensor1 = tf.reshape(t1, shape=(2, 256))
    tensor2 = tf.reshape(tensor2, shape=(256,))
    return tensor1, tensor2

def construct_datatset_numpy(args1):
    global args
    args = args1
    data1 = tf.random.uniform((15000, 2, 256), maxval=128, dtype=tf.dtypes.int64)
    segs = tf.zeros((15000, 256), dtype=tf.dtypes.int64)

    train_dataset = tf.data.Dataset.from_tensor_slices((data1, segs))
    val_dataset = tf.data.Dataset.from_tensor_slices((data1, segs))
    
    train_dataset = train_dataset.map(lambda t1, t2: tf.py_function(func=test_map_numpy,
        inp=[t1, t2], Tout=(tf.int64, tf.int64)))
    val_dataset = val_dataset.map(lambda t1, t2: tf.py_function(func=test_map_numpy,
        inp=[t1, t2], Tout=(tf.int64, tf.int64)))

    train_dataset = train_dataset.prefetch(tf.data.experimental.AUTOTUNE)
    val_dataset = val_dataset.prefetch(tf.data.experimental.AUTOTUNE)

    train_dataset = train_dataset.batch(args.batch_size, drop_remainder=True)
    val_dataset = val_dataset.batch(args.batch_size, drop_remainder=True)

    # for x, y in train_dataset.take(2):
    #     print(x.shape)
    return train_dataset, val_dataset

def generator_ids(tokenizer_ro, tokenizer_bert, args):

    with open(args.dataset_file, 'r', encoding='utf-8', errors='replace') as f:
        for i, line in enumerate(f):
            if i % 2 == 0:
                target = line.strip()
            elif i % 2 == 1:
                source = line.strip()
                (source, target), segments = encode_gec(source, target, tokenizer_ro, tokenizer_bert, args)
                if len(source) > args.seq_length or len(target) > args.seq_length:
                    continue
                yield (source, target), segments

def gec_generator_text(args):
    with open(args.dataset_file, 'r', encoding='utf-8', errors='replace') as f:
        for i, line in enumerate(f):
            if i % 2 == 0:
                target = line.strip()
            elif i % 2 == 1:
                source = line.strip()
                yield (source, target)

def make_fixed_length(input_ids: List[int], max_seq_len: int):

    if len(input_ids) < max_seq_len: # pad
        to_add = max_seq_len-len(input_ids)
        for _ in range(to_add):
            input_ids.append(0)
    elif len(input_ids) > max_seq_len: # trim
        input_ids = input_ids[:max_seq_len]
    return input_ids

def encode_gec(source: str, target: str, tokenizer_ro, tokenizer_bert, args):

    if args.bert:
        tokens = ['[CLS]']
        tokens.extend(tokenizer_bert.tokenize(source))
        tokens.append('[SEP]')
        source = tokenizer_bert.convert_tokens_to_ids(tokens)
    else:
        source = [tokenizer_ro.vocab_size] + tokenizer_ro.encode(source) +\
            [tokenizer_ro.vocab_size + 1]
    target = [tokenizer_ro.vocab_size] + tokenizer_ro.encode(target) +\
            [tokenizer_ro.vocab_size + 1]

    segments = [0] * len(source) + [1] * (args.seq_length - len(source))
    source = make_fixed_length(source, args.seq_length)
    target = make_fixed_length(target, args.seq_length)
    segments = make_fixed_length(segments, args.seq_length)

    return (source, target), segments

def generator_tensors_ids():
    global tokenizer_bert, tokenizer_ro, args
    gen = generator_ids(tokenizer_ro, tokenizer_bert, args)
    for step, ((s, t), segs) in enumerate(gen):
        if step % 10000 == 0:
            tf.compat.v1.logging.info('tf record generation, step {}'.format(step))
        s = tf.convert_to_tensor(s, dtype=tf.int64)
        t = tf.convert_to_tensor(t, dtype=tf.int64)
        segs = tf.convert_to_tensor(segs, dtype=tf.int64)
        # print(s.shape, t.shape, segs.shape, tf.stack([s, t]).shape)
        yield tf.stack([s, t]), segs

def generator_tensors_ids_and_segs():
    global tokenizer_bert, tokenizer_ro, args

    gen = generator_ids(tokenizer_ro, tokenizer_bert, args)
    for step, ((s, t), segs) in enumerate(gen):
        if step % 10000 == 0:
            tf.compat.v1.logging.info('tf record generation, step {}'.format(step))
        
        yield (tf.convert_to_tensor(s, dtype=tf.int64), tf.convert_to_tensor(t, dtype=tf.int64)),\
                tf.convert_to_tensor(segs, dtype=tf.int64)
                

def get_text_samples(args) -> List[str]:
    gen = gec_generator_text(args)
    return gen

def construct_tokenizer(examples: List, subwords_path, args):

    if examples is None:
        tokenizer_ro = tfds.features.text.SubwordTextEncoder.load_from_file(subwords_path)
        return tokenizer_ro

    merged_examples = []
    for (source, target) in examples:
        merged_examples.append(source)
        merged_examples.append(target)

    tokenizer_ro = tfds.features.text.SubwordTextEncoder.build_from_corpus(
            merged_examples, target_vocab_size=args.dict_size)
    
    if not os.path.exists(subwords_path):
        os.makedirs(subwords_path)

    tokenizer_ro.save_to_file(subwords_path)

    return tokenizer_ro



from typing import List, Callable, Dict
import csv
from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType
from rb.complexity.word.valence_type import ValenceTypeEnum
from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class Valence(ComplexityIndex):


    valence_dict: Dict[str, Dict[str, float]] = None

    def __init__(self, lang: Lang, valence_type: ValenceTypeEnum,
                 reduce_depth: int, reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.WORD,
                                 abbr="Valence", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)
        if Valence.valence_dict is None:
            Valence.parse_valence_csv(lang)
        self.valence_type = valence_type

    @staticmethod
    def parse_valence_csv(lang: Lang):

        if lang is Lang.EN:
            path_to_valence = 'resources/en/wordlists/valences_en_with_liwc.csv'
        Valence.valence_dict = {}
        for vt in ValenceTypeEnum:
            Valence.valence_dict[vt] = {}

        with open(path_to_valence) as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=';')
            for i, row in enumerate(csv_reader):
                if i < 2:
                    continue
                row[0] = row[0].strip()                    
                for j, vt in enumerate(ValenceTypeEnum):
                    if len(str(row[j])) > 0 and str(row[j + 1]) != '0':
                        Valence.valence_dict[vt][row[0]] = float(row[j + 1])

    def _compute_value(self, element: TextElement) -> List[float]:
        return sum(1 for word in element.get_words() if word.text in Valence.valence_dict[self.valence_type])
    
    def __repr__(self):
        return f"{self.reduce_function_abbr}({self.abbr}_{self.valence_type.name.lower()} / {self.reduce_depth_abbr})"


import os
from rb.core.word import Word
from typing import Dict, List

import numpy as np
import tensorflow as tf
from rb import Block, Document
from rb.core.lang import Lang
from rb.similarity.vector import Vector
from rb.similarity.vector_model import VectorModel, VectorModelType
from rb.utils.rblogger import Logger
from transformers import AutoConfig, AutoTokenizer, TFAutoModel

logger = Logger.get_logger()


class TransformersEncoder(VectorModel):

    def __init__(self, lang: Lang, name: str = None, max_seq_len=512):
        from_pt = False
        if name is None:
            if lang is Lang.EN:
                name = "roberta-base"
                self.SOW = "Ġ"
            elif lang is Lang.FR:
                name = "camembert-base"
                self.SOW = "▁"
            elif lang is Lang.RO:
                name = "readerbench/RoBERT-base"
                self.SOW = "##"
            elif lang is Lang.RU:
                name = "DeepPavlov/rubert-base-cased"
                from_pt=True
                self.SOW = "##"
        VectorModel.__init__(self, VectorModelType.TRANSFORMER, name=name, lang=lang, size=768)  
        config = AutoConfig.from_pretrained(self.name)
        config.output_hidden_states = True
        self.tokenizer = AutoTokenizer.from_pretrained(self.name)
        self.bert = TFAutoModel.from_pretrained(self.name, config=config, from_pt=from_pt)
        self.max_seq_len = max_seq_len
        self.bos = self.tokenizer.bos_token_id if self.tokenizer.bos_token_id is not None else self.tokenizer.cls_token_id
        self.eos = self.tokenizer.eos_token_id if self.tokenizer.eos_token_id is not None else self.tokenizer.sep_token_id

    @staticmethod
    def common_prefix(word1: str, word2: str) -> str:
        i = 0
        while i < min(len(word1), len(word2)):
            if word1[i] != word2[i]:
                return word1[:i]
            i += 1
        return word1[:i]

    def _create_word_token_dict(self, block: Block, tokenized: List[str]) -> Dict[Word, List[int]]:
        i = 0
        result = {}
        tokens = [token.replace(self.SOW, "") for token in tokenized]
        current = ""
        for word in block.get_words():
            text = word.text
            if getattr(self.tokenizer, "do_lower_case", False):
                text = text.lower()
            while not tokens[i]:
                i += 1
            current = self.common_prefix(text, tokens[i])
            if not current:
                continue 
            ids = [i]
            if text == current:
                if len(text) < len(tokens[i]):
                    tokens[i] = tokens[i][len(current):]
                else:
                    i += 1
            else: 
                i += 1
                while i < len(tokens) and text.startswith(current + tokens[i]):
                    ids.append(i)
                    current += tokens[i]
                    i += 1
                current = text[len(current):]
                if len(current) > 0:
                    tokens[i] = tokens[i][len(current):]
            result[word] = ids
        return result

    def _encode_block(self, block: Block):
        tokenized = self.tokenizer.tokenize(block.text)
        token_index = self._create_word_token_dict(block, tokenized)
        start = 0
        n = len(tokenized)
        token_ids = self.tokenizer.convert_tokens_to_ids(tokenized)
        embeddings = []
        while start < n:
            end = min(n, start + self.max_seq_len - 2)
            input_ids = np.array([[self.bos] + token_ids[start:end] + [self.eos]])
            attention_mask = np.ones(input_ids.shape, dtype=np.int32)
            outputs = self.bert(input_ids=input_ids, attention_mask=attention_mask)
            embeddings.append(outputs.hidden_states[-2][0, 1:-1, :].numpy())
            start = end
        embeddings = np.concatenate(embeddings, axis=0)
        for sentence in block.get_sentences():
            for word in sentence.get_words():
                if word in token_index:
                    word.vectors[self] = Vector(np.mean(embeddings[token_index[word]], axis=0))
            sentence_ids = [
                i
                for word in sentence.get_words() 
                if word in token_index 
                for i in token_index[word]
            ]
            if sentence_ids:
                sentence.vectors[self] = Vector(np.mean(embeddings[sentence_ids], axis=0))
        block.vectors[self] = Vector(np.mean(embeddings, axis=0))

    def encode(self, document: Document):
        vectors = []
        for block in document.get_blocks():
            self._encode_block(block)
            if self in block.vectors:
                vectors.append(block.vectors[self].values)
        if vectors:
            document.vectors[self] = Vector(np.mean(vectors, axis=0))